# -*- coding: utf-8 -*-
"""
Created on Wed Apr 14 18:34:29 2021

@author: amarmore

Module used to handle compute the Barwise TF matrix, presented in [1]
(Barwise TF matrix: a 2D representation of barwise features, 
each feature representing Time-Frequency content, where time is expressed at barscale)

See [1 - Chapter 2.4] or [2] for more information.

References
----------
[1] Marmoret, A. (2022). Unsupervised Machine Learning Paradigms for the Representation of Music Similarity and Structure (Doctoral dissertation, Université Rennes 1).
https://theses.hal.science/tel-04589687

[2] Marmoret, A., Cohen, J.E, and Bimbot, F., "Barwise Compression Schemes 
for Audio-Based Music Structure Analysis"", in: 19th Sound and Music Computing Conference, 
SMC 2022, Sound and music Computing network, 2022.
"""

import as_seg.data_manipulation as dm
import as_seg.model.errors as err

import numpy as np
import tensorly as tl
import librosa

# %% Spectrograms to tensors
# !!! Be extremely careful with the organization of modes, which can be either Frequency-Time at barscale-Bars (FTB) or Bars-Frequency-Time at barscale (BFT) depending on the method.
def spectrogram_to_tensor_barwise(spectrogram, bars, hop_length_seconds, subdivision, mode_order="BFT", subset_nb_bars = None):
    """
    Spectrogram to tensor-spectrogram, with the order of modes defined by the mode_order parameter.
    """
    if mode_order == "BFT":
        return tensorize_barwise_BFT(spectrogram, bars, hop_length_seconds, subdivision, subset_nb_bars)
    
    elif mode_order == "FTB":
        return tensorize_barwise_FTB(spectrogram, bars, hop_length_seconds, subdivision, subset_nb_bars)
    
    else:
        raise err.InvalidArgumentValueException(f"Unknown mode order: {mode_order}.")

def tensorize_barwise_BFT(spectrogram, bars, hop_length_seconds, subdivision, subset_nb_bars = None):
    """
    Returns a 3rd order tensor-spectrogram from the original spectrogram and bars starts and ends.
    The order of modes is Bars-Frequency-Time at barscale (BFT).
    Must be used for SSAE and the computtion of Barwise TF matrix.
    
    Each bar in the tensor-spectrogram contains the same number of frames, define by the "subdivision" parameter.
    These frames are selected from an oversampled spectrogram, adapting to the specific size of each bar.
    See [1] for details.

    Parameters
    ----------
    spectrogram : list of list of floats or numpy array
        The spectrogram to return as a tensor-spectrogram.
    bars : list of tuples
        List of the bars (start, end), in seconds, to cut the spectrogram at bar delimitation.
    hop_length_seconds : float
        The hop_length, in seconds.
    subdivision : integer
        The number of subdivision of the bar to be contained in each slice of the tensor.

    Returns
    -------
    np.array tensor
        The tensor-spectrogram as a np.array.

    """
    barwise_spec = []
    bars_idx = dm.segments_from_time_to_frame_idx(bars[1:], hop_length_seconds)
    if subset_nb_bars is not None:
        bars_idx = bars_idx[:subset_nb_bars]
    for idx, beats in enumerate(bars_idx):
        t_0 = beats[0]
        t_1 = beats[1]
        samples = [int(round(t_0 + k * (t_1 - t_0)/subdivision)) for k in range(subdivision)]
        if len(samples) != len(set(samples)): # Check for repetitions
            if idx != len(bars_idx) - 1: # It's not a problem if it's the last bar, because it is due to inconsistencies between the last downbeat estimated and the end of the song.
                raise err.ToDebugException("The subdivision is too large, it leads to repeated samples chosen in the bar!")
        if samples[-1] < spectrogram.shape[1]:
            barwise_spec.append(spectrogram[:,samples])
    return np.array(barwise_spec)

def tensorize_barwise_FTB(spectrogram, bars, hop_length_seconds, subdivision, subset_nb_bars = None):
    #(careful: different mode organization than previous one: here, this is Frequency-Time-Bars)
    """
    Returns a 3rd order tensor-spectrogram from the original spectrogram and bars starts and ends.
    The order of modes is Frequency-Time at barscale-Bars (FTB).
    Must be used for NTD.
    
    Each bar in the tensor-spectrogram contains the same number of frames, define by the "subdivision" parameter.
    These frames are selected from an oversampled spectrogram, adapting to the specific size of each bar.
    See [1, Chap 2.4.2] for details.

    Parameters
    ----------
    spectrogram : list of list of floats or numpy array
        The spectrogram to return as a tensor-spectrogram.
    bars : list of tuples
        List of the bars (start, end), in seconds, to cut the spectrogram at bar delimitation.
    hop_length_seconds : float
        The hop_length, in seconds.
    subdivision : integer
        The number of subdivision of the bar to be contained in each slice of the tensor.

    Returns
    -------
    tensorly tensor
        The tensor-spectrogram as a tensorly tensor.

    """
    freq_len = spectrogram.shape[0]
    bars_idx = dm.segments_from_time_to_frame_idx(bars[1:], hop_length_seconds)
    if subset_nb_bars is not None:
        bars_idx = bars_idx[:subset_nb_bars]
    samples_init = [int(round(bars_idx[0][0] + k * (bars_idx[0][1] - bars_idx[0][0])/subdivision)) for k in range(subdivision)]

    tens = np.array(spectrogram[:,samples_init]).reshape(freq_len, subdivision, 1)
    
    for bar in bars_idx[1:]:
        t_0 = bar[0]
        t_1 = bar[1]
        samples = [int(round(t_0 + k * (t_1 - t_0)/subdivision)) for k in range(subdivision)]
        if samples[-1] < spectrogram.shape[1]:
            current_bar_tensor_spectrogram = spectrogram[:,samples].reshape(freq_len, subdivision,1)
            tens = np.append(tens, current_bar_tensor_spectrogram, axis = 2)
        else:
            break
    
    return tl.tensor(tens)#, dtype=tl.float32)

# %% Tensors to spectrograms
def tensor_barwise_to_spectrogram(tensor, mode_order = "BFT", subset_nb_bars = None):
    """
    Return a spectrogram from a tensor-spectrogram, with the order of modes defined by the mode_order parameter.
    """
    if subset_nb_bars is not None:
        tensor = barwise_subset_this_tensor(tensor, subset_nb_bars, mode_order = mode_order)
    
    if mode_order == "BFT":
        return tl.unfold(tensor, 1)
    
    elif mode_order == "FTB":
        return np.reshape(tensor, (tensor.shape[0], tensor.shape[1] * tensor.shape[2]), order = 'F') # Note: it is NOT the same than unfold(0)
    
    else:
        raise err.InvalidArgumentValueException(f"Unknown mode order: {mode_order}.")

def barwise_subset_this_tensor(tensor, subset_nb_bars, mode_order = "BFT"):
    """
    Keep only the subset_nb_bars first bars in the tensor.
    """
    if mode_order == "BFT":
        return tensor[:subset_nb_bars]
   
    elif mode_order == "FTB":
        return tensor[:,:,:subset_nb_bars]

    else:
        raise err.InvalidArgumentValueException(f"Unknown mode order: {mode_order}.")
    
def get_this_bar_tensor(tensor, bar_idx, mode_order = "BFT"):
    """
    Return one particular bar of the tensor.
    """
    if mode_order == "BFT":
        return tensor[bar_idx]
   
    elif mode_order == "FTB":
        return tensor[:,:,bar_idx]

    else:
        raise err.InvalidArgumentValueException(f"Unknown mode order: {mode_order}.")

# %% Spectrogram to Barwise TF matrix
def barwise_TF_matrix(spectrogram, bars, hop_length_seconds, subdivision, subset_nb_bars = None):
    """
    Barwise TF matrix, a 2D representation of Barwise spectrograms as Time-Frequency vectors.
    See [1] for details.

    Parameters
    ----------
    spectrogram : list of list of floats or numpy array
        The spectrogram to return as a tensor-spectrogram.
    bars : list of tuples
        List of the bars (start, end), in seconds, to cut the spectrogram at bar delimitation.
    hop_length_seconds : float
        The hop_length, in seconds.
    subdivision : integer
        The number of subdivision of the bar to be contained in each slice of the tensor.

    Returns
    -------
    np.array
        The Barwise TF matrix, of sizes (b, tf).

    """
    tensor_spectrogram = tensorize_barwise_BFT(spectrogram, bars, hop_length_seconds, subdivision, subset_nb_bars=subset_nb_bars)
    return tl.unfold(tensor_spectrogram, 0)

def barwise_subset_this_TF_matrix(matrix, subset_nb_bars):
    """
    Keep only the subset_nb_bars first bars in the Barwise TF matrix.
    """
    assert subset_nb_bars is not None
    return matrix[:subset_nb_bars]

# %% Vector and Barwise TF to spectrogram
def TF_vector_to_spectrogram(vector, frequency_dimension, subdivision):
    """
    Encapsulating the conversion from a Time-Frequency vector to a Time-Frequency matrix (spectrogram)

    Parameters
    ----------
    vector : np.array
        A Time-Frequency vector (typically a row in the Barwise TF matrix).
    frequency_dimension : positive integer
        The size of the frequency dimension 
        (number of components in this dimension).
    subdivision : positive integer
        The size of the time dimension at the bar scale 
        (number of time components in each bar, defined as parameter when creating the Barwise TF matrix).

    Returns
    -------
    np.array
        A Time-Frequency matrix (spectrogram) of size (frequency_dimension, subdivision).

    """
    assert frequency_dimension*subdivision == vector.shape[0]
    return tl.fold(vector, 0, (frequency_dimension,subdivision))

def TF_matrix_to_spectrogram(matrix, frequency_dimension, subdivision, subset_nb_bars = None):
    """
    Encapsulating the conversion from a Barwise TF matrix to a spectrogram.
    """
    spectrogram_content = None
    if subset_nb_bars is not None:
        matrix = barwise_subset_this_TF_matrix(matrix, subset_nb_bars)
    for tf_vector in matrix:
        bar_content = TF_vector_to_spectrogram(tf_vector, frequency_dimension, subdivision)
        spectrogram_content = np.concatenate((spectrogram_content, bar_content), axis=1) if spectrogram_content is not None else bar_content
    return spectrogram_content


# Tensor to Barwise TF
def tensor_barwise_to_barwise_TF(tensor, mode_order = "BFT"):
    """
    Return the Barwise TF matrix from a tensor-spectrogram, with the order of modes defined by the mode_order parameter.
    """
    # Barmode: 0 for BTF, 2 for FTB
    if mode_order == "BFT":
        return tl.unfold(tensor, 0)
    elif mode_order == "FTB":
        return tl.unfold(tensor, 2)
    else:
        raise err.InvalidArgumentValueException(f"Unknown mode order: {mode_order}.")

# %% Barwise TF to tensor
# TODO

# Beatwise MSAF
def beat_synchronize_msaf(spectrogram, frame_times, beat_frames, beat_times):
    # Make beat synchronous
    beatsync_feats = librosa.util.utils.sync(spectrogram.T, beat_frames, pad=True).T

    # Assign times (and add last time if padded)
    beatsync_times = np.copy(beat_times)
    if beatsync_times.shape[0] != beatsync_feats.shape[0]:
        beatsync_times = np.concatenate((beatsync_times,
                                         [frame_times[-1]]))
    return beatsync_feats, beatsync_times



# -*- coding: utf-8 -*-
"""
Created on Tue Feb 25 16:01:35 2020

@author: amarmore

Module defining specific errors to be raised
"""

class ArgumentException(BaseException): pass
class InvalidArgumentValueException(ArgumentException): pass
class NegativeValuesNotAccepted(InvalidArgumentValueException): pass

class AbnormalBehaviorException(BaseException): pass
class ToDebugException(AbnormalBehaviorException): pass
class BuggedFunctionException(AbnormalBehaviorException): pass
class OutdatedBehaviorException(AbnormalBehaviorException): pass





# -*- coding: utf-8 -*-
"""
Created on Fri Feb 22 16:29:17 2019

@author: amarmore

Defining common plotting functions.
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm

from base_audio.common_plot import *

# %% Plotting utils
def plot_lenghts_hist(lengths):
    """
    Plots the lengths of segments in an histogram
    i.e. the distribution of the size of segments in the annotation/estimation 
    (allegedly already computed into a list lengths).
 
    Parameters
    ----------
    lengths : list of integers
        List of all segments' sizes in the annotation/estimation.
    """
    plt.rcParams.update({'font.size': 18})

    fig, axs = plt.subplots(1, 1, figsize=(6, 3.75))
    axs.hist(lengths, bins = range(1,34), density = True, cumulative = False, align = "left")
    plt.xticks(np.concatenate([[1],range(4, 34, 4)]))
    plt.ylim(0,1)

    axs.set_xlabel("Size of the segment,\nin number of bars")
    axs.set_ylabel("Proportion among\nall segments")

    plt.show()

def plot_measure_with_annotations(measure, annotations, color = "red"):
    """
    Plots the measure (typically novelty) with the segmentation annotation.
    """
    plt.plot(np.arange(len(measure)),measure, color = "black")
    for x in annotations:
        plt.plot([x, x], [0,np.amax(measure)], '-', linewidth=1, color = color)
    plt.show()
    
def plot_measure_with_annotations_and_prediction(measure, annotations, frontiers_predictions, title = "Title"):
    """
    Plots the measure (typically novelty) with the segmentation annotation and the estimated segmentation.
    """
    plt.title(title)
    plt.plot(np.arange(len(measure)),measure, color = "black")
    ax1 = plt.axes()
    ax1.axes.get_yaxis().set_visible(False)
    for x in annotations:
        plt.plot([x, x], [0,np.amax(measure)], '-', linewidth=1, color = "red")
    for x in frontiers_predictions:
        if x in annotations:
            plt.plot([x, x], [0,np.amax(measure)], '-', linewidth=1, color = "#8080FF")#"#17becf")
        else:
            plt.plot([x, x], [0,np.amax(measure)], '-', linewidth=1, color = "orange")
    plt.show()
    
def plot_segments_with_annotations(seg, annot):
    """
    Plots the estimated labelling of segments next to with the frontiers in the annotation.
    """
    for x in seg:
        plt.plot([x[0], x[1]], [x[2]/10,x[2]/10], '-', linewidth=1, color = "black")
    for x in annot:
        plt.plot([x[1], x[1]], [0,np.amax(np.array(seg)[:,2])/10], '-', linewidth=1, color = "red")
    plt.show()
    


# -*- coding: utf-8 -*-
"""
Created on Wed Nov 15 2023

@author: amarmore

A test sample for the CBM algorithm.
"""

# External imports
# Module for manipulating arrays
import numpy as np

# Module for reading signals
import librosa

# Module to handle annotations for MIR files
import mirdata

# Module encapsulating the computation of features from the signal
import as_seg.model.signal_to_spectrogram as signal_to_spectrogram

# General module for manipulating data: conversion between time, bars, frame indexes, loading of data, ...
import as_seg.data_manipulation as dm

# Module to process the input in barwise features
import as_seg.barwise_input as bi

# Module to process the compute the autosimilarity
import as_seg.autosimilarity_computation as as_computation

# Module containing the CBM algorithm
import as_seg.CBM_algorithm as cbm

# Plotting module
from as_seg.model.common_plot import *

# %% Loading annotations and defining the audio path
path_to_beatles_dataset = '/home/a23marmo/datasets/beatles' # To change
beatles = mirdata.initialize('beatles', path_to_beatles_dataset)
beatles.download()

# NB: you have to place the file "01_-_Come_Together.wav" manually in the folder "audio/11_-_Abbey_Road/"
come_together = beatles.track('1101')

references_segments = beatles.load_sections(come_together.sections_path).intervals

song_path = come_together.audio_path

# %% Loading the song, as signal, and estimate downbeats
the_signal, sampling_rate = librosa.load(song_path, sr = None)
bars = dm.get_bars_from_audio(song_path)

# %% Computing the features
hop_length = 32 # Oversampling the spectrogram, to select frames which will be equally-spaced barwise.
hop_length_seconds = hop_length/sampling_rate # As bars are in seconds, we convert this hop length in seconds.
subdivision_bars = 96 # The number of time samples to consider in each bar.

feature_object = signal_to_spectrogram.FeatureObject(sr=sampling_rate, feature="log_mel", hop_length=hop_length, mel_grill=True)
log_mel = feature_object.get_spectrogram(the_signal) # Log_mel spectrogram

barwise_TF = bi.barwise_TF_matrix(log_mel, bars, hop_length_seconds, subdivision_bars)

# %% Cosine autosimilarity
barwise_TF_cosine_autosimilarity = as_computation.switch_autosimilarity(barwise_TF, "cosine")
#Alternatively, one could use: as_computation.get_cosine_autosimilarity(barwise_TF_cosine)
plot_me_this_spectrogram(barwise_TF_cosine_autosimilarity, title = "Cosine autosimilarity of the Barwise TF matrix")

#%% Running the CBM on the autosimilarity matrix
segments_cbm_cosine = cbm.compute_cbm(barwise_TF_cosine_autosimilarity, penalty_weight = 1, penalty_func = "modulo8", bands_number = 7)[0]
segments_cbm_cosine_in_time = dm.segments_from_bar_to_time(segments_cbm_cosine, bars)

score_cbm_cosine_zero_five = dm.compute_score_of_segmentation(references_segments, segments_cbm_cosine_in_time, window_length = 0.5)
print(f"Score with 0.5 second tolerance: Precision {score_cbm_cosine_zero_five[0]}, Recall {score_cbm_cosine_zero_five[1]}, F measure {score_cbm_cosine_zero_five[2]}")
score_cbm_cosine_three = dm.compute_score_of_segmentation(references_segments, segments_cbm_cosine_in_time, window_length = 3)
print(f"Score with 3 seconds tolerance: Precision {score_cbm_cosine_three[0]}, Recall {score_cbm_cosine_three[1]}, F measure {score_cbm_cosine_three[2]}")

# %% Autocorrelation/Covariance autosimilarity
barwise_TF_covariance_autosimilarity = as_computation.switch_autosimilarity(barwise_TF, "covariance")
plot_me_this_spectrogram(barwise_TF_covariance_autosimilarity, title = "Covariance autosimilarity of the Barwise TF matrix")

# %% Running the CBM on the autosimilarity matrix
segments_cbm_covariance = cbm.compute_cbm(barwise_TF_covariance_autosimilarity, penalty_weight = 1, penalty_func = "modulo8", bands_number = 7)[0]
segments_cbm_covariance_in_time = dm.segments_from_bar_to_time(segments_cbm_covariance, bars)

score_cbm_covariance_zero_five = dm.compute_score_of_segmentation(references_segments, segments_cbm_covariance_in_time, window_length = 0.5)
print(f"Score with 0.5 second tolerance: Precision {score_cbm_covariance_zero_five[0]}, Recall {score_cbm_covariance_zero_five[1]}, F measure {score_cbm_covariance_zero_five[2]}")
score_cbm_covariance_three = dm.compute_score_of_segmentation(references_segments, segments_cbm_covariance_in_time, window_length = 3)
print(f"Score with 3 seconds tolerance: Precision {score_cbm_covariance_three[0]}, Recall {score_cbm_covariance_three[1]}, F measure {score_cbm_covariance_three[2]}")

# %% RBF autosimilarity
barwise_TF_rbf_autosimilarity = as_computation.switch_autosimilarity(barwise_TF, "RBF")
plot_me_this_spectrogram(barwise_TF_rbf_autosimilarity, title = "RBF autosimilarity of the Barwise TF matrix")

# %% Running the CBM on the autosimilarity matrix
segments_cbm_rbf = cbm.compute_cbm(barwise_TF_rbf_autosimilarity, penalty_weight = 1, penalty_func = "modulo8", bands_number = 7)[0]
segments_cbm_rbf_in_time = dm.segments_from_bar_to_time(segments_cbm_rbf, bars)

score_cbm_rbf_zero_five = dm.compute_score_of_segmentation(references_segments, segments_cbm_rbf_in_time, window_length = 0.5)
print(f"Score with 0.5 second tolerance: Precision {score_cbm_rbf_zero_five[0]}, Recall {score_cbm_rbf_zero_five[1]}, F measure {score_cbm_rbf_zero_five[2]}")
score_cbm_rbf_three = dm.compute_score_of_segmentation(references_segments, segments_cbm_rbf_in_time, window_length = 3)
print(f"Score with 3 seconds tolerance: Precision {score_cbm_rbf_three[0]}, Recall {score_cbm_rbf_three[1]}, F measure {score_cbm_rbf_three[2]}")

# %% Conclusion
import pandas as pd
columns = np.array(["Precision 0.5", "Recall 0.5", "F measure 0.5","Precision 3", "Recall 3", "F measure 3"])  
tab = []
tab.append([round(score_cbm_cosine_zero_five[0],5), round(score_cbm_cosine_zero_five[1],5),round(score_cbm_cosine_zero_five[2],5),round(score_cbm_cosine_three[0],5),round(score_cbm_cosine_three[1],5),round(score_cbm_cosine_three[2],5)])
tab.append([round(score_cbm_covariance_zero_five[0],5), round(score_cbm_covariance_zero_five[1],5),round(score_cbm_covariance_zero_five[2],5),round(score_cbm_covariance_three[0],5),round(score_cbm_covariance_three[1],5),round(score_cbm_covariance_three[2],5)])
tab.append([round(score_cbm_rbf_zero_five[0],5), round(score_cbm_rbf_zero_five[1],5),round(score_cbm_rbf_zero_five[2],5),round(score_cbm_rbf_three[0],5),round(score_cbm_rbf_three[1],5),round(score_cbm_rbf_three[2],5)])

pd.DataFrame(tab, index=["Cosine", "Covariance", "RBF"], columns=columns)


# -*- coding: utf-8 -*-
"""
Created on Mon Mar 14 16:30:31 2022

@author: amarmore

Module used to compute autosimilarity matrices.
"""
import as_seg.model.errors as err

import numpy as np
import sklearn.metrics.pairwise as pairwise_distances
import warnings
eps = 1e-10

def switch_autosimilarity(an_array, similarity_type, gamma = None, normalise = True):
    """
    High-level function to compute the autosimilarity of this matrix.
    
    Expects a matrix of shape (Bars, Feature representation).
    
    Computes it with different possible similarity function s_{x_i,x_j} (given two bars denoted as x_i and x_j):
        - "cosine" for the cosine similarity, i.e. the normalised dot product:
        .. math::
            s_{x_i,x_j} = \\frac{\langle x_i, x_j \rangle}{||x_i|| ||x_j||}
        -"covariance" for a covariance similarity, 
        i.e. the dot product of centered features:
        .. math::
            s_{x_i,x_j} = \langle x_i - \hat{x}, x_j - \hat{x} \rangle
        -"rbf" for the Radial Basis Function similarity, 
        i.e. the exponent of the opposite of the euclidean distance between features:
        .. math::
            s_{x_i,x_j} = \\exp^{-\\gamma ||x_i - x_j||_2}
        The euclidean distance can be the distance between the normalised features.
        Gamma is a parameter.
        See rbf_kernel from scikit-learn for more details.
    
    Parameters
    ----------
    an_array : numpy array
        The array/matrix seen as array which autosimilarity will be computed.
        Expected to be of shape (Bars, Feature representation).
    similarity_type : string
        Either "cosine", "covariance" or "rbf".
        It represents the similarity function used for computing the autosimilarity.
    gamma : positive float, optional
        The gamma parameter in the rbf function, only used for the "rbf" similarity.
        The default is None, meaning that it is computed as function of the standard deviation,
        see get_gamma_std() for more details.
    normalise : boolean, optional
        Whether features should be normalised or not. 
        Normalisation depends on the similarity function.
        The default is True.

    Returns
    -------
    numpy array
        Autosimilarity matrix of the input an_array.

    """
    if similarity_type.lower() == "cosine":
        return get_cosine_autosimilarity(an_array)#, normalise = normalise)
    elif similarity_type.lower() == "autocorrelation" or similarity_type.lower() == "covariance":
        return get_autocorrelation_autosimilarity(an_array, normalise = normalise)
    elif similarity_type.lower() == "rbf":
        return get_rbf_autosimilarity(an_array, gamma, normalise = normalise)
    elif similarity_type.lower() == "centered_rbf":
        return get_centered_rbf_autosimilarity(an_array, gamma, normalise = normalise)
    else:
        raise err.InvalidArgumentValueException(f"Incorrect similarity type: {similarity_type}. Should be cosine, covariance or rbf.")
        
def l2_normalise_barwise(an_array):
    """
    Normalises the array barwise (i.e., in its first dimension) by the l_2 norm.
    
    Null values are replaced by the small positive value of 10^{-10}.

    Parameters
    ----------
    an_array : numpy array
        The array which needs to be normalised.

    Returns
    -------
    numpy array
        The normalised array.

    """
    norm = np.linalg.norm(an_array, axis = 1)
    an_array_T = np.transpose(an_array)
    out = np.inf * np.ones_like(an_array_T)
    np.divide(an_array_T, norm, out = out, where=norm!=0)
    an_array_T = np.where(np.isinf(out), eps, out)
    return np.transpose(an_array_T)

def get_cosine_autosimilarity(an_array):#, normalise = True):
    """
    Computes the autosimilarity matrix with the cosine similarity function.
    
    The cosine similarity function is the normalised dot product between two bars, i.e.:
    .. math::
        s_{x_i,x_j} = \\frac{\langle x_i, x_j \rangle}{||x_i|| ||x_j||}
    
    Parameters
    ----------
    an_array : numpy array
        The array/matrix seen as array which autosimilarity os to compute.
        Expected to be of shape (Bars, Feature representation).

    Returns
    -------
    numpy array
        The autosimilarity of this array, with the cosine similarity function.

    """
    if type(an_array) is list:
        this_array = np.array(an_array)
    else:
        this_array = an_array
    #if normalise:
    this_array = l2_normalise_barwise(this_array)
    return this_array@this_array.T

def get_covariance_autosimilarity(an_array, normalise = True):
    """
    Note: deprecated. The name of the matrix became "Autocorrelation" in the TISMIR version of the paper.
    """
    return get_autocorrelation_autosimilarity(an_array, normalise = normalise)

def get_autocorrelation_autosimilarity(an_array, normalise = True):
    """
    Computes the autosimilarity matrix, where the similarity function is the autocorrelation.
    
    The autocorrelation similarity function corresponds to the dot product of centered features:
    .. math::
        s_{x_i,x_j} = \langle x_i - \hat{x}, x_j - \hat{x} \rangle

    Parameters
    ----------
    an_array : numpy array
        The array/matrix seen as array which autosimilarity will be computed.
    normalise : boolean, optional
        Whether features should be normalised or not. 
        Normalisation here means that each centered feature is normalised by its norm.
        The default is True.
        
    Returns
    -------
    numpy array
        The autocorrelation autosimilarity of this array.

    """
    if type(an_array) is list:
        this_array = np.array(an_array)
    else:
        this_array = an_array
    this_array = this_array - this_array.mean(axis=0) # centering, i.e. subtracting the average value row-wise
    if normalise:
        this_array = l2_normalise_barwise(this_array)
    return this_array@this_array.T

def get_rbf_autosimilarity(an_array, gamma = None, normalise = True):
    """
    Computes the autosimilarity matrix, where the similarity function is the Radial Basis Function (RBF).
    
    The RBF corresponds to the exponent of the opposite of the euclidean distance between features:
    .. math::
        s_{x_i,x_j} = \\exp^{-\\gamma ||x_i - x_j||_2}
        
    The RBF is computed via scikit-learn.
    The default gamma value is computed in function get_gamma_std(), refer to that function for further details.

    Parameters
    ----------
    an_array : numpy array
        The array/matrix seen as array which autosimilarity will be computed.
    gamma : positive float, optional
        The gamma parameter in the rbf function.
        The default is None, meaning that it is computed as function of the standard deviation,
        see get_gamma_std() for more details.
    normalise : boolean, optional
        Whether features should be normalised or not. 
        Normalisation here means that the euclidean norm is computed between normalised vectors.
        The default is True.

    Returns
    -------
    numpy array
        The RBF autosimilarity of this array.

    """
    if type(an_array) is list:
        this_array = np.array(an_array)
    else:
        this_array = an_array
    if gamma == None:
        gamma = get_gamma_std(this_array, scaling_factor = 1, no_diag = True, normalise = normalise)
    if normalise:
        this_array = l2_normalise_barwise(this_array)
    return pairwise_distances.rbf_kernel(this_array, gamma = gamma)
    
def get_centered_rbf_autosimilarity(an_array, gamma = None, normalise = True):
    """
    TODO
    """
    if type(an_array) is list:
        this_array = np.array(an_array)
    else:
        this_array = an_array
    this_array = this_array - this_array.mean(axis=0) # centering, i.e. subtracting the average value row-wise
    if gamma == None:
        gamma = get_gamma_std(this_array, scaling_factor = 1, no_diag = True, normalise = normalise)
    if normalise:
        this_array = l2_normalise_barwise(this_array)
    return pairwise_distances.rbf_kernel(this_array, gamma = gamma)


def get_gamma_std(an_array, scaling_factor = 1, no_diag = True, normalise = True):
    """
    Default value for the gamma in the RBF similarity function.
    
    This default value is proportional to the inverse of the standard deviation of the values, more experiments should be made to fit it.
    For now, it has been set quite empirically.

    Parameters
    ----------
    an_array : numpy array
        The array/matrix seen as array which autosimilarity will be computed.
    scaling_factor : positive float, optional
        Weigthing parameter, relating to the inverse of the standard deviation. 
        The default is 1.
    no_diag : boolen, optional
        Whether the diagonal values (self similarity values) should be discarded (True) or taken into account (False). 
        The default is True.
    normalise : boolean, optional
        Whether features should be normalised or not. 
        Normalisation here means that the euclidean norm is computed between normalised vectors.
        The default is True.

    Returns
    -------
    gamma : float
        The gamma parameter in the RBF similarity function.

    """
    if normalise:
        an_array = l2_normalise_barwise(an_array)
    euc_dist = pairwise_distances.euclidean_distances(an_array)
    if not no_diag:
        return scaling_factor/(2*np.std(euc_dist))
    else:
        for i in range(len(euc_dist)):
            euc_dist[i,i] = float('NaN')
        return scaling_factor/(2*np.nanstd(euc_dist))


# -*- coding: utf-8 -*-
"""
Created on Wed Mar 25 16:54:59 2020

@author: amarmore

Computing spectrogram in different feature description.

It is now moved to base_audio package.
"""

from base_audio.signal_to_spectrogram import *


# -*- coding: utf-8 -*-
"""
Created on Mon Feb 24 11:01:42 2020

@author: amarmore

Convolutive "Block-Matching" (CBM) algorithm.
This algorithm is designed to segment barwise autosimilarities.

In short, this algorithm focuses on the criteria of homogeneity to estimate segments, 
and computes an optimal segmentation via dynamic programming.

See [1] for more details.

References
----------
[1] Marmoret, A., Cohen, J. E., & Bimbot, F. (2023). Barwise Music Structure Analysis with the Correlation Block-Matching Segmentation Algorithm. Transactions of the International Society for Music Information Retrieval (TISMIR), 6(1), 167-185.
"""

import as_seg.model.errors as err

import math
import numpy as np
from scipy.sparse import diags
import warnings

def compute_cbm(autosimilarity, min_size = 1, max_size = 32, penalty_weight = 1, penalty_func = "modulo8", bands_number = None):
    """
    Dynamic programming algorithm, maximizing an overall score at the song scale, sum of all segments' scores on the autosimilarity.
    Each segment' score is a combination of
     - the convolution score on this the segment, dependong on the kernel, 
     - a penalty cost, function of the size of the segment, to enforce specific sizes (with prior knowledge),

    The penalty cost is computed in the function "penalty_cost_from_arg()".
    See this function for further details.

    It returns the optimal segmentation according to this score.
    
    This algortihm is also described in [1].
    
    IDEAS FOR FUTURE DEVELOPMENT: 
        - May be optimized using scipy.signal.fftconvolve(), but requires a different approach in parsing all segments.
    (i.e. parsing all possible segments for a specified kernel size and then retrieve the indexes of the diagonally-centered values)
        - Taking into account values which are not around the diagonal but everywhere in the matrix in order to account for the repetition principle.

    Parameters
    ----------
    autosimilarity : list of list of float (list of columns)
        The autosimilarity to segment.
    min_size : integer, optional
        The minimal length of segments.
        The default is 1.
    max_size : integer, optional
        The maximal length of segments.
        The default is 32.
    penalty_weight : float, optional
        The ponderation parameter for the penalty function
    penalty_func : string
        The type of penalty function to use.
        See "penalty_cost_from_arg()" for further details.
    bands_number : positive integer or None, optional
        The number of bands in the kernel. 
        For the full kernel, bands_number must be set to None
        (or higher than the maximal size, but cumbersome)
        See [1] for details. 
        The default is None.

    Raises
    ------
    ToDebugException
        If the program fails, generally meaning that the autosimilarity is incorrect.

    Returns
    -------
    list of tuples
        The segments, as a list of tuples (start, end).
    integer
        Global cost (the maximum among all).
    """
    scores = [-math.inf for i in range(len(autosimilarity))]
    segments_best_starts = [None for i in range(len(autosimilarity))]
    segments_best_starts[0] = 0
    scores[0] = 0
    kernels = compute_all_kernels(max_size, bands_number = bands_number)
    max_conv_eight = np.amax(convolution_entire_matrix_computation(autosimilarity, kernels))
    
    for current_idx in range(1, len(autosimilarity)): # Parse all indexes of the autosimilarity
        for possible_start_idx in possible_segment_start(current_idx, min_size = min_size, max_size = max_size):
            if possible_start_idx < 0:
                raise err.ToDebugException("Invalid value of start index, shouldn't happen.") from None
                
            # Convolutionnal cost between the possible start of the segment and the current index (entire segment)
            conv_cost = convolutionnal_cost(autosimilarity[possible_start_idx:current_idx,possible_start_idx:current_idx], kernels)
                        
            segment_length = current_idx - possible_start_idx
            penalty_cost = penalty_cost_from_arg(penalty_func, segment_length)            
            
            this_segment_cost = conv_cost * segment_length - penalty_cost * penalty_weight * max_conv_eight
            # Note: conv_eight is not normalized by its size (not a problem in itself as size is contant, but generally not specified in formulas).

            if possible_start_idx == 0: # Avoiding errors, as scores values are initially set to -inf.
                if this_segment_cost > scores[current_idx]: # This segment is of larger score
                    scores[current_idx] = this_segment_cost
                    segments_best_starts[current_idx] = 0
            else:
                if scores[possible_start_idx] + this_segment_cost > scores[current_idx]: # This segment is of larger score
                    scores[current_idx] = scores[possible_start_idx] + this_segment_cost
                    segments_best_starts[current_idx] = possible_start_idx

    segments = [(segments_best_starts[len(autosimilarity) - 1], len(autosimilarity) - 1)]
    precedent_frontier = segments_best_starts[len(autosimilarity) - 1] # Because a segment's start is the previous one's end.
    while precedent_frontier > 0:
        segments.append((segments_best_starts[precedent_frontier], precedent_frontier))
        precedent_frontier = segments_best_starts[precedent_frontier]
        if precedent_frontier == None:
            raise err.ToDebugException("Well... The dynamic programming algorithm took an impossible path, so it failed. Understand why.") from None
    return segments[::-1], scores[-1]

def compute_all_kernels(max_size, bands_number = None):
    """
    Precomputes all kernels of size 0 ([0]) to max_size, to be reused in the CBM algorithm.
    
    This is used for acceleration purposes.

    Parameters
    ----------
    max_size : integer
        The maximal size (included) for kernels.
    bands_number : positive integer or None, optional
        The number of bands in the kernel. 
        For the full kernel, bands_number must be set to None
        (or higher than the maximal size, but cumbersome)
        See [1] for details. 
        The default is None.
        
    Returns
    -------
    kernels : array of arrays (which are kernels)
        All the kernels, of size 0 ([0]) to max_size.

    """
    kernels = [[0]]
    for p in range(1,max_size + 1):
        if bands_number is None or p < bands_number:
            kern = np.ones((p,p)) - np.identity(p)
        else:
            k = np.array([np.ones(p-i) for i in np.abs(range(-bands_number, bands_number + 1))],dtype=object)
            offset = [i for i in range(-bands_number, bands_number + 1)]
            kern = diags(k,offset).toarray() - np.identity(p)
        kernels.append(kern)
    return kernels

def convolutionnal_cost(cropped_autosimilarity, kernels):
    """
    The convolution measure on this part of the autosimilarity matrix.

    Parameters
    ----------
    cropped_autosimilarity : list of list of floats or numpy array (matrix representation)
        The part of the autosimilarity which convolution measure is to compute.
    kernels : list of arrays
        Acceptable kernels.

    Returns
    -------
    float
        The convolution measure.

    """
    p = len(cropped_autosimilarity)
    kern = kernels[p]
    #return np.mean(np.multiply(kern,cropped_autosimilarity))
    return np.sum(np.multiply(kern,cropped_autosimilarity)) / p**2
    
def convolution_entire_matrix_computation(autosimilarity_array, kernels, kernel_size = 8):
    """
    Computes the convolution measure on the entire autosimilarity matrix, with a defined and fixed kernel size.
    
    Parameters
    ----------
    autosimilarity_array : list of list of floats or numpy array (matrix representation)
        The autosimilarity matrix.
    kernels : list of arrays
        All acceptable kernels.
    kernel_size : integer
        The size of the kernel for this measure.

    Returns
    -------
    cost : list of float
        List of convolution measures, at each bar of the autosimilarity.

    """
    cost = np.zeros(len(autosimilarity_array))
    for i in range(kernel_size, len(autosimilarity_array)):
        cost[i] = convolutionnal_cost(autosimilarity_array[i - kernel_size:i,i - kernel_size:i], kernels)
    return cost

def penalty_cost_from_arg(penalty_func, segment_length):
    """
    Returns a penalty cost, function of the size of the segment.
    The penalty function has to be specified, and is bound to evolve in the near future,
    so this docstring won't explain it.
    Instead, you'll have to read the code, sorry! It is pretty straightforward though.
    
    The ``modulo'' functions are based on empirical prior knowledge,
    following the fact that pop music is generally composed of segments of 4 or 8 bars.
    Still, penalty values are empirically set.

    Parameters
    ----------
    penalty_func : string
        Identifier of the penalty function.
    segment_length : integer
        Size of the segment.

    Returns
    -------
    float
        The penalty cost.

    """
    if penalty_func == "modulo8":        
        if segment_length == 8:
            return 0
        elif segment_length %4 == 0:
            return 1/4
        elif segment_length %2 == 0:
            return 1/2
        else:
            return 1
    if penalty_func == "modulo4":
        if segment_length %4 == 0:
            return 0
        elif segment_length %2 == 0:
            return 1/2
        else:
            return 1
    if penalty_func == "modulo8modulo4":        
        if segment_length == 8:
            return 0
        elif segment_length == 4:
            return 1/4
        elif segment_length %2 == 0:
            return 1/2
        else:
            return 1
    if penalty_func == "target_deviation_8_alpha_half": 
         return abs(segment_length - 8) ** (1/2)
    if penalty_func == "target_deviation_8_alpha_one": 
         return abs(segment_length - 8)
    if penalty_func == "target_deviation_8_alpha_two": 
         return abs(segment_length - 8) ** 2
    else:
        raise err.InvalidArgumentValueException(f"Penalty function not understood {penalty_func}.")

def possible_segment_start(idx, min_size = 1, max_size = None):
    """
    Generates the list of all possible starts of segments given the index of its end.
    
    Parameters
    ----------
    idx: integer
        The end of a segment.
    min_size: integer
        Minimal length of a segment.
    max_size: integer
        Maximal length of a segment.
        
    Returns
    -------
    list of integers
        All potentials starts of structural segments.
    """
    if min_size < 1: # No segment should be allowed to be 0 size
        raise err.InvalidArgumentValueException(f"Invalid minimal size: {min_size} (No segment should be allowed to be 0 or negative size).")
        #min_size = 1
    if max_size == None:
        return range(0, idx - min_size + 1)
    else:
        if idx >= max_size:
            return range(idx - max_size, idx - min_size + 1)
        elif idx >= min_size:
            return range(0, idx - min_size + 1)
        else:
            return []
        
# %% Scikit-learn class
# Author: Axel Marmoret
# 
# Adapted from: https://scikit-learn.org/stable/auto_examples/developing_estimators/sklearn_is_fitted.html, Author: Kushan <kushansharma1@gmail.com>
#
# License: BSD 3 clause

from sklearn.base import BaseEstimator, ClassifierMixin

import as_seg.autosimilarity_computation as as_comp
import as_seg.data_manipulation as dm

class CBMEstimator(BaseEstimator, ClassifierMixin):
    """
    Scikit-learn class for the CBM algorithm. May be used for practicity, following the scikit-learn API.
    """
    def __init__(self, similarity_function="cosine", max_size=32, penalty_weight=1, penalty_func="modulo8", bands_number=7):
        """
        Constructor of the CBM estimator.

        Parameters
        ----------
        similarity_function : string, optional
            The similarity function to use for computing the autosimilarity.
            The default is "cosine".
        max_size : integer, optional
            The maximal size of segments.
            The default is 32.
        penalty_weight : float, optional
            The ponderation parameter for the penalty function.
            The default is 1.
        penalty_func : string, optional
            The type of penalty function to use.
            The default is "modulo8".
        bands_number : positive integer or None, optional
            The number of bands in the kernel.
            For the full kernel, bands_number must be set to None
            (or higher than the maximal size, but cumbersome)
            See [1] for details.
            The default is 7.
        """
        self.similarity_function = similarity_function
        self.max_size = max_size
        self.penalty_weight = penalty_weight
        self.penalty_func = penalty_func
        self.bands_number = bands_number
        self.algorithm_name = "CBM"

    def predict(self, barwise_features):
        """
        Perform Predictions

        If the estimator is not fitted, then raise NotFittedError
        """
        ssm_matrix = as_comp.switch_autosimilarity(barwise_features, similarity_type=self.similarity_function)
        segments = compute_cbm(ssm_matrix, max_size=self.max_size, penalty_weight=self.penalty_weight, 
                               penalty_func=self.penalty_func, bands_number = self.bands_number)[0]
        return segments
    
    def predict_in_seconds(self, barwise_features, bars):
        """
        Perform Predictions, and convert the segments from bars to seconds.

        If the estimator is not fitted, then raise NotFittedError
        """
        segments = self.predict(barwise_features)
        return dm.segments_from_bar_to_time(segments, bars)
    
    def predict_in_seconds_this_autosimilarity(self, ssm_matrix, bars):
        """
        Perform Predictions on a given autosimilarity matrix, and convert the segments from bars to seconds.

        If the estimator is not fitted, then raise NotFittedError
        """
        segments = compute_cbm(ssm_matrix, max_size=self.max_size, penalty_weight=self.penalty_weight, 
                               penalty_func=self.penalty_func, bands_number = self.bands_number)[0]
        return dm.segments_from_bar_to_time(segments, bars)

    def score(self, predictions, annotations):
        """
        Compute the score of the predictions
        """
        close_tolerance = dm.compute_score_of_segmentation(annotations, predictions, window_length=0.5)
        large_tolerance = dm.compute_score_of_segmentation(annotations, predictions, window_length=3)
        return close_tolerance, large_tolerance


from . import autosimilarity_computation
from . import barwise_input
from . import data_manipulation
from . import CBM_algorithm

from .model import common_plot
from .model import dataloaders
from .model import errors
from .model import signal_to_spectrogram



# -*- coding: utf-8 -*-
"""
Created on Wed Sep 22 17:22:59 2021

Default paths for the different folders.
Should be changed prior to any computation.

@author: amarmore
"""
import os

# We suppose that we are in the Notebooks folder,
# hence data is in the parent folder.
path_parent_of_data = os.path.dirname(os.getcwd())

# Global
path_data_persisted = f"{path_parent_of_data}/data/data_persisted"

# RWC
#path_data_persisted_rwcpop = "C:/Users/amarmore/Desktop/data_persisted/RWC Pop" ## Path where pre-computed data on RWC Pop should be stored (bars, beats, spectrograms, ssm, ...)
path_annotation_rwcpop = f"{path_parent_of_data}/data/annotations/rwcpop" ## Path of the annotations of RWC Pop. Should be parent folder of "AIST" and/or "MIREX10"
path_entire_rwcpop = "C:/Users/amarmore/Desktop/Audio samples/RWC Pop/Entire RWC" ## Path where are stored wav files of RWC
#path_even_songs_rwcpop = "C:/Users/amarmore/Desktop/Audio samples/RWC Pop/Even songs" ## Path containing only the wav files of songs of odd numbers
#path_odd_songs_rwcpop = "C:/Users/amarmore/Desktop/Audio samples/RWC Pop/Odd songs" ## Path containing only the wav files of songs of even numbers
#path_debug_rwcpop = "C:/Users/amarmore/Desktop/Audio samples/RWC Pop/debug" ## Debug path, containing only two songs, to fix functions quickly
#path_mirdata_rwcpop = "C:/Users/amarmore/Desktop/Audio samples/mirdata/RWC Pop/"

# SALAMI
path_data_persisted_salami = f"{path_parent_of_data}/data/annotations/salami" ## Path where pre-computed data on the SALAMI dataset should be stored (bars, beats, spectrograms, ssm, ...)
path_entire_salami = "C:/Users/amarmore/Desktop/Audio samples/SALAMI" ## Path where are stored wav files of SALAMI (path where it is downloaded by mirdata also)

# Come Together
come_together = "/home/a23marmo/this_folder/The Beatles - Come Together"
path_data_persisted_come_together = "/home/a23marmo/Bureau/data_persisted/cometogether"


"""
Script to display experimental results in a nice format.

I should probably use scikit-learn's GridSearchCV, TODO.
"""

import pandas as pd
from IPython.display import display
import numpy as np
import itertools

def display_experimental_results(data, conditions, metrics, emphasis = None):
    assert data.shape[-1] == len(metrics), f"Wrong number of metrics: {data.shape[-1]} != {len(metrics)}"
    if type(conditions) == list:
        conditions = np.array(conditions, dtype=object)
    if len(conditions.shape) == 1 and type(conditions[0]) not in [list, tuple, np.ndarray]:
        indexes = conditions
    elif len(conditions.shape) > 1 or (len(conditions.shape) == 1 and type(conditions) in [list, tuple, np.ndarray]):
        indexes = np.array(list(itertools.product(*conditions))) # cartesian_product
    else:
        raise NotImplementedError(f"Wrong format of conditions: {conditions}")
    
    dataframe = pd.DataFrame(data, columns = metrics, index = indexes)
    if emphasis is not None:
        display(dataframe.style.bar(subset=emphasis, color='#5fba7d'))
    else:
        display(dataframe)
    
def find_best_condition(data, conditions):
    if type(conditions) == list:
        conditions = np.array(conditions, dtype=object)
    if len(conditions.shape) == 1 and type(conditions[0]) not in [list, tuple, np.ndarray]:
        best_arg = np.argmax(data)
        best_val = conditions[best_arg]
        return best_val
    elif len(conditions.shape) > 1 or (len(conditions.shape) == 1 and type(conditions) in [list, tuple, np.ndarray]):
        best_arg_flatten = np.argmax(data)
        best_arg_tuple = np.unravel_index(best_arg_flatten, data.shape)
        best_vals = []
        for idx, best_arg in enumerate(best_arg_tuple):
            best_vals.append(conditions[idx][best_arg])
        return np.array(best_vals)

        




# -*- coding: utf-8 -*-
"""
Created on Wed Mar  4 17:34:39 2020

@author: amarmore

A module containing some high-level scripts for decomposition and/or segmentation.
Not meant to be shared but rather to work.
"""

import librosa.core
import librosa.feature
import librosa
import os
import numpy as np

import as_seg.data_manipulation as dm
import as_seg.barwise_input as bi
import as_seg.model.signal_to_spectrogram as signal_to_spectrogram
import as_seg.model.errors as err
import as_seg.scripts.default_path as paths


# %% SALAMI test dataset
def get_salami_test_indexes():
    list_songs_in_test = open(f"{paths.path_data_persisted_salami}/test_set_salami_gs.txt")
    test_dataset = []
    for part in list_songs_in_test.readlines():
        line_broken = part.split("\n")
        test_dataset.append(int(line_broken[0]))
    return test_dataset

def is_in_salami_train(key, salami_test_dataset = None):
    if salami_test_dataset is None: # May be passed as argument to avoid loading the file at each test.
        salami_test_dataset = get_salami_test_indexes()
    return not key in salami_test_dataset

def is_in_salami_test(key, salami_test_dataset = None):
    if salami_test_dataset is None: # May be passed as argument to avoid loading the file at each test.
        salami_test_dataset = get_salami_test_indexes()
    return key in salami_test_dataset

# %% Loading (or saving) multiple data
def load_spec_annot_song_RWC(song_number, feature, hop_length = 32):
    """
    Load the spectrogram, the bar segmentation and the annotation for a given song in RWC Pop.
    For this function to work, paths should be updated to correct ones.

    Parameters
    ----------
    song_number : integer
        Index of the song in the RWC Pop dataset (1 to 100).
    feature : string
        The feature description to compute the spectrogram with.
    hop_length : int, optional
        Hop length in the computation of the spetrogram.
        The default is 32.

    Returns
    -------
    spectrogram : numpy array
        The spectorgram of this song, searched in memory.
    bars : list of tuple of floats
        The bars for this song, searched in memory.
    references_segments : np.array
        Segments of reference (annotations) for this song.
    """
    spectrogram = load_spectrogram("rwcpop", song_number, feature, hop_length, fmin = 98, n_fft = 2048, n_mfcc = 20)
    bars = load_bars("rwcpop", song_number)
    annotations_mirex = f"{paths.path_annotation_rwc}/MIREX10"
    annot_path_mirex = "{}/{}".format(annotations_mirex, dm.get_annotation_name_from_song(song_number, "MIREX10"))
    annotations = dm.get_segmentation_from_txt(annot_path_mirex, "MIREX10")
    references_segments = np.array(annotations)[:,0:2]
    return spectrogram, bars, references_segments

def load_or_save_spec_annot_song_RWC(song_number, feature, hop_length = 32):
    """
    Load the spectrogram, the bar segmentation and the annotation for a given song in RWC Pop.
    For this function to work, paths should be updated to correct ones,
    and outputs will be stored if not computed (spectrograms and bars typically).

    Parameters
    ----------
    song_number : integer
        Index of the song in the RWC Pop dataset (1 to 100).
    feature : string
        The feature description to compute the spectrogram with.
    hop_length : int, optional
        Hop length in the computation of the spetrogram.
        The default is 32.

    Returns
    -------
    spectrogram : numpy array
        The spectorgram of this song, either computed or found in memory.
    bars : list of tuple of floats
        The bars for this song, either computed or found in memory.
    references_segments : np.array
        Segments of reference (annotations) for this song.
    """
    spectrogram = load_or_save_spectrogram("rwcpop", f"{paths.path_entire_rwc}/{song_number}.wav", feature, hop_length, fmin = 98, n_fft = 2048, n_mfcc = 20)
    bars = load_or_save_bars("rwcpop", f"{paths.path_entire_rwc}/{song_number}.wav")
    annotations_mirex = f"{paths.path_annotation_rwc}/MIREX10"
    annot_path_mirex = "{}/{}".format(annotations_mirex, dm.get_annotation_name_from_song(song_number, "MIREX10"))
    annotations = dm.get_segmentation_from_txt(annot_path_mirex, "MIREX10")
    references_segments = np.array(annotations)[:,0:2]
    return spectrogram, bars, references_segments

def load_bar_annot_song_RWC(song_number):
    """
    Similar to load_spec_annot_song_RWC(), but without loading the soectrogram 
    (only bars and annotations).
    
    Generally used when spectrogram is not desired but has a huge memory cost (or may not be known).
    """
    bars = load_bars("rwcpop", song_number)
    annotations_mirex = f"{paths.path_annotation_rwcpop}/MIREX10"
    annot_path_mirex = "{}/{}".format(annotations_mirex, dm.get_annotation_name_from_song(song_number, "MIREX10"))
    annotations = dm.get_segmentation_from_txt(annot_path_mirex, "MIREX10")
    references_segments = np.array(annotations)[:,0:2]
    return bars, references_segments

def load_or_save_bar_annot_song_RWC(song_number):
    """
    Similar to load_or_save_spec_annot_song_RWC(), but without the spectrogram 
    (only bars and annotations).
    
    Generally used when spectrogram is not desired but has a huge computation or memory cost.
    """
    bars = load_or_save_bars("rwcpop", f"{paths.path_entire_rwc}/{song_number}.wav")
    annotations_mirex = f"{paths.path_annotation_rwc}/MIREX10"
    annot_path_mirex = "{}/{}".format(annotations_mirex, dm.get_annotation_name_from_song(song_number, "MIREX10"))
    annotations = dm.get_segmentation_from_txt(annot_path_mirex, "MIREX10")
    references_segments = np.array(annotations)[:,0:2]
    return bars, references_segments

def load_beat_annot_song_RWC(song_number):
    """
    Loads beats and annotations for this song.
    """
    beats = load_beats("rwcpop", song_number)
    annotations_mirex = f"{paths.path_annotation_rwcpop}/MIREX10"
    annot_path_mirex = "{}/{}".format(annotations_mirex, dm.get_annotation_name_from_song(song_number, "MIREX10"))
    annotations = dm.get_segmentation_from_txt(annot_path_mirex, "MIREX10")
    references_segments = np.array(annotations)[:,0:2]
    return beats, references_segments

def load_or_save_spectrogram_and_bars(dataset, song_path, feature, hop_length, fmin = 98, n_fft = 2048, n_mfcc = 20):
    """
    Loads the spectrogram and the bars for this song, which were persisted after a first computation, or compute them if they weren't found.

    Parameters
    ----------
    dataset : string
        Name of the dataset. Only "rwcpop" and "salami" for now.
    song_path : string
        The path of the signal of the song.
    feature : string
        Feature of the spectrogram, part of the identifier of the spectrogram.
    hop_length : integer
        hop_length of the spectrogram, part of the identifier of the spectrogram.
    fmin : integer
        Minimal frequence for the spectrogram, part of the identifier of the spectrogram.
        The default is 98.
    n_fft and n_mfcc : integers, optional
        Both arguments are used respectively for the stft and for the mfcc computation, and are used to 

    Returns
    -------
    bars : list of tuple of floats
        The persisted bars for this song.
    spectrogram : numpy array
        The pre-computed spectorgram.
    """
    bars = load_or_save_bars(dataset, song_path)
    spectrogram = load_or_save_spectrogram(dataset, song_path, feature, hop_length, fmin = fmin, n_fft = n_fft, n_mfcc = n_mfcc)
    return bars, spectrogram

def load_or_save_bars_and_barwise_tf(dataset, song_path, feature = "log_mel_grill", subdivision = 96, hop_length=32):
    bars = load_or_save_bars(dataset, song_path)
    barwise_tf = load_or_save_barwise_tf(dataset, song_path, bars, feature, subdivision, hop_length)
    return bars, barwise_tf

def load_or_save_beats_and_beatwise_tf(dataset, song_path, feature = "log_mel_grill", subdivision_beat = 24, hop_length=32):
    beats = load_or_save_beats(dataset, song_path)
    beatwise_tf = load_or_save_beatwise_tf(dataset, song_path, beats, feature, subdivision_beat, hop_length)
    return beats, beatwise_tf

def load_spec_annot_cometogether(feature, hop_length = 32):
    """
    Load the spectrogram, the bar segmentation and the annotation for Come Together, from the Beatles.
    For this function to work, paths should be updated to correct ones,
    and outputs will be stored if not computed (spectrograms and bars typically).

    Parameters
    ----------
    feature : string
        The feature description to compute the spectrogram with.
    hop_length : int, optional
        Hop length in the computation of the spetrogram.
        The default is 32.

    Returns
    -------
    spectrogram : numpy array
        The spectorgram of this song, either computed or found in memory.
    bars : list of tuple of floats
        The bars for this song, either computed or found in memory.
    references_segments : np.array
        Segments of reference (annotations) for this song.
    """
    spectrogram = load_spectrogram("cometogether", "come_together", feature, hop_length, fmin = 98, n_fft = 2048, n_mfcc = 20)
    bars = load_bars("cometogether", f"{paths.come_together}.wav")
    annotation_path = f"{paths.come_together}.lab"
    annotations = dm.get_segmentation_from_txt(annotation_path, "MIREX10")
    references_segments = np.array(annotations)[:,0:2]
    return spectrogram, bars, references_segments

def load_or_save_spec_annot_cometogether(feature, hop_length = 32):
    """
    Load the spectrogram, the bar segmentation and the annotation for Come Together, from the Beatles.
    For this function to work, paths should be updated to correct ones,
    and outputs will be stored if not computed (spectrograms and bars typically).

    Parameters
    ----------
    feature : string
        The feature description to compute the spectrogram with.
    hop_length : int, optional
        Hop length in the computation of the spetrogram.
        The default is 32.

    Returns
    -------
    spectrogram : numpy array
        The spectorgram of this song, either computed or found in memory.
    bars : list of tuple of floats
        The bars for this song, either computed or found in memory.
    references_segments : np.array
        Segments of reference (annotations) for this song.
    """
    spectrogram = load_or_save_spectrogram("cometogether", f"{paths.come_together}.wav", feature, hop_length, fmin = 98, n_fft = 2048, n_mfcc = 20)
    bars = load_or_save_bars("cometogether", f"{paths.come_together}.wav")
    annotation_path = f"{paths.come_together}.lab"
    annotations = dm.get_segmentation_from_txt(annotation_path, "MIREX10")
    references_segments = np.array(annotations)[:,0:2]
    return spectrogram, bars, references_segments

# %% Loading (or saving) spectrograms
def load_spectrogram(dataset, song_name, feature, hop_length, fmin = 98, n_fft = 2048, n_mfcc = 20):
    """
    Lloads the spectrogram for this song (supposing it was already computed).

    Parameters
    ----------
    dataset : string
        Name of the dataset (should be either "rwcpop" or "salami")
    song_name : string
        The name of the song.
    feature : string
        Feature of the spectrogram, part of the identifier of the spectrogram.
    hop_length : integer
        hop_length of the spectrogram, part of the identifier of the spectrogram.
    fmin : integer
        Minimal frequence for the spectrogram, part of the identifier of the spectrogram.
        The default is 98.

    Returns
    -------=
    spectrogram : numpy array
        The pre-computed spectorgram.
    """
    name_spectrogram = format_name_spectrogram_params(song_name, feature, hop_length, fmin, n_fft, n_mfcc)
    return np.load(f"{paths.path_data_persisted}/{dataset}/spectrograms/{name_spectrogram}.npy")

def load_or_save_spectrogram(dataset, song_path, feature, hop_length, fmin = 98, n_fft = 2048, n_mfcc = 20):
    """
    Computes the spectrogram for this song, or load it if it were already computed.

    Parameters
    ----------
    persisted_path : string
        Path where the spectrogram should be found.
    song_path : string
        The path of the signal of the song.
    feature : string
        Feature of the spectrogram, part of the identifier of the spectrogram.
    hop_length : integer
        hop_length of the spectrogram, part of the identifier of the spectrogram.
    fmin : integer
        Minimal frequence for the spectrogram, part of the identifier of the spectrogram.
        The default is 98.

    Returns
    -------=
    spectrogram : numpy array
        The pre-computed spectorgram.
    """
    song_name = song_path.split("/")[-1].replace(".wav","").replace(".mp3","")
    try:
        return load_spectrogram(dataset, song_name, feature, hop_length, fmin, n_fft, n_mfcc)
    except FileNotFoundError:
        if "log_mel_grill" in feature:
            mel = load_or_save_spectrogram(dataset, song_path, "mel_grill", hop_length, fmin = fmin, n_fft = n_fft, n_mfcc = n_mfcc)
            return signal_to_spectrogram.get_log_mel_from_mel(mel, feature)
        elif feature == "mel" or feature == "log_mel" or feature == "logmel":
            raise err.InvalidArgumentValueException(f"Invalid mel parameter ({feature}), are't you looking for mel_grill?")
        else:
            the_signal, _ = librosa.load(song_path, sr=44100)
            if "stft" in feature:
                if "nfft" not in feature: 
                    spectrogram = signal_to_spectrogram.get_spectrogram(the_signal, 44100, "stft", hop_length, n_fft = n_fft)
                else:              
                    n_fft_arg = int(feature.split("nfft")[1])
                    spectrogram = signal_to_spectrogram.get_spectrogram(the_signal, 44100, "stft", hop_length, n_fft = n_fft_arg)
            elif "mfcc" in feature:
                if "nmfcc" not in feature:
                    spectrogram = signal_to_spectrogram.get_spectrogram(the_signal, 44100, "mfcc", hop_length, n_mfcc = n_mfcc)   
                else:
                    n_mfcc_arg = int(feature.split("nmfcc")[1])
                    spectrogram = signal_to_spectrogram.get_spectrogram(the_signal, 44100, "mfcc", hop_length, n_mfcc = n_mfcc_arg)
            elif feature == "pcp_tonnetz":
                # If chromas are already computed, try to load them instead of recomputing them.
                chromas = load_or_save_spectrogram(dataset, song_path, "pcp", hop_length, fmin = fmin)
                spectrogram = librosa.feature.tonnetz(y=None, sr = None, chroma = chromas)
            elif feature == "pcp":
                # If it wasn't pcp_tonnetz, compute the spectrogram, and then save it.
                spectrogram = signal_to_spectrogram.get_spectrogram(the_signal, 44100, feature, hop_length, fmin = fmin)
            else:
                spectrogram = signal_to_spectrogram.get_spectrogram(the_signal, 44100, feature, hop_length)
        
            name_spectrogram = format_name_spectrogram_params(song_name, feature, hop_length, fmin, n_fft, n_mfcc)
            np.save(f"{paths.path_data_persisted}/{dataset}/spectrograms/{name_spectrogram}", spectrogram)
            return spectrogram

def format_name_spectrogram_params(song_name, feature, hop_length, fmin, n_fft, n_mfcc):
    if "stft" in feature:   
        if "nfft" not in feature:
            return f"{song_name}_{feature}-nfft{n_fft}_stereo_{hop_length}"
        else:
            return f"{song_name}_{feature}_stereo_{hop_length}"
    elif "mfcc" in feature:
        if "nmfcc" not in feature:
            return f"{song_name}_{feature}-nmfcc{n_mfcc}_stereo_{hop_length}"
        else:
            return f"{song_name}_{feature}_stereo_{hop_length}"
    elif feature == "pcp":
        return f"{song_name}_{feature}_stereo_{hop_length}_{fmin}"
    elif feature == "mel" or feature == "log_mel":
        raise err.InvalidArgumentValueException("Invalid mel parameter, are't you looking for mel_grill?")
    else:
        return f"{song_name}_{feature}_stereo_{hop_length}"
    
# %% Loading (or saving) bars and beats
def load_bars(dataset, song_name):
    """
    Loads the bars for this song, which were persisted after a first computation.

    Parameters
    ----------
    dataset : string
        Name of the dataset. For now are only handled "salami" and "rwcpop".
    song_name : string
        Name of the song (identifier of the bars to load).

    Returns
    -------
    bars : list of tuple of floats
        The persisted bars for this song.
    """
    return np.load(f"{paths.path_data_persisted}/{dataset}/bars/{song_name}.npy")

def load_or_save_bars(dataset, song_path):
    """
    Computes the bars for this song, or load them if they were already computed.

    Parameters
    ----------
    dataset : string
        Name of the dataset. Only "rwcpop" and "salami" for now.
    song_path : string
        The path of the signal of the song.

    Returns
    -------
    bars : list of tuple of floats
        The persisted bars for this song.
    """
    song_name = song_path.split("/")[-1].replace(".wav","").replace(".mp3","")
    try:
        bars = load_bars(dataset, song_name)
    except:
        bars = dm.get_bars_from_audio(song_path)
        np.save(f"{paths.path_data_persisted}/{dataset}/bars/{song_name}", bars)
    return bars

def load_beats(dataset, song_name):
    """
    Loads the beats for this song, which were persisted after a first computation.

    Parameters
    ----------
    dataset : string
        Name of the dataset. For now are only handled "salami" and "rwcpop".
    song_name : string
        Name of the song (identifier of the bars to load).

    Returns
    -------
    beats : list of tuple of floats
        The persisted beats for this song.
    """
    return np.load(f"{paths.path_data_persisted}/{dataset}/beats/{song_name}.npy")

def load_or_save_beats(dataset, song_path):
    """
    Computes the beats for this song, or load them if they were already computed.

    Parameters
    ----------
    dataset : string
        Name of the dataset. Only "rwcpop" and "salami" for now.
    song_path : string
        The path of the signal of the song.

    Returns
    -------
    beats : list of tuple of floats
        The persisted beats for this song.
    """
    song_name = song_path.split("/")[-1].replace(".wav","").replace(".mp3","")
    try:
        beats = load_beats(dataset, song_name)
    except:
        beats = dm.get_beats_from_audio_madmom(song_path)
        np.save(f"{paths.path_data_persisted}/{dataset}/beats/{song_name}", beats)
    return beats

# %% load (or save) TF matrices
def get_name_tf_matrices(dataset, song_name, feature, subdivision):
    return f"{dataset}_{song_name}_{feature}_subdiv{subdivision}"

def load_barwise_tf(dataset, song_name, feature = "log_mel_grill", subdivision = 96):
    """
    Load the barwiseTF matrix (already pre-computed) for this song.
    """
    name_data = get_name_tf_matrices(dataset, song_name, feature, subdivision)
    return np.load(f"{paths.path_data_persisted}/{dataset}/barwise_tf/{name_data}.npy", allow_pickle = True)

def load_or_save_barwise_tf(dataset, song_path, bars, feature = "log_mel_grill", subdivision = 96, hop_length = 32):
    """
    Load the barwiseTF matrix (already pre-computed) for this song.
    """
    song_name = song_path.split("/")[-1].replace(".wav","").replace(".mp3","")
    try:
        return load_barwise_tf(dataset, song_name, feature, subdivision)
    
    except FileNotFoundError:
        spectrogram = load_or_save_spectrogram(dataset, song_path, feature, hop_length = hop_length)
        barwise_tf = bi.barwise_TF_matrix(spectrogram, bars, hop_length/44100, subdivision)
        name_data = get_name_tf_matrices(dataset, song_name, feature, subdivision)
        np.save(f"{paths.path_data_persisted}/{dataset}/barwise_tf/{name_data}", barwise_tf)
        return barwise_tf

def load_beatwise_tf(dataset, song_name, feature = "log_mel_grill", subdivision_beat = 24):
    """
    Load the beatwiseTF matrix (already pre-computed) for this song.
    """
    name_data = get_name_tf_matrices(dataset, song_name, feature, subdivision_beat)
    return np.load(f"{paths.path_data_persisted}/{dataset}/beatwise_tf/{name_data}.npy", allow_pickle = True)

def load_or_save_beatwise_tf(dataset, song_path, beats, feature = "log_mel_grill", subdivision_beat = 24, hop_length=32):
    song_name = song_path.split("/")[-1].replace(".wav","").replace(".mp3","")
    try:
        return load_beatwise_tf(dataset, song_name, feature, subdivision_beat)
    
    except FileNotFoundError:
        spectrogram = load_or_save_spectrogram(dataset, song_path, feature, hop_length = hop_length)
        beatwise_tf = bi.barwise_TF_matrix(spectrogram, beats, hop_length/44100, subdivision_beat)
        name_data = get_name_tf_matrices(dataset, song_name, feature, subdivision_beat)
        np.save(f"{paths.path_data_persisted}/{dataset}/beatwise_tf/{name_data}", beatwise_tf)
        return beatwise_tf
    
# %% Loading SSM - useful for TISMIR recomputation
def load_barwise_tf_ssm(dataset, song_name, feature = "log_mel_grill", subdivision = 96, similarity_type = "cosine", train = False):
    if dataset == "salami":
        if train:
            subset = "train"
        else:
            subset = "test"
        return np.load(f"{paths.path_data_persisted}/salami/self_similarity_matrices/{subset}/ssm_{dataset}_song{song_name}_barwiseTF_{similarity_type}_{feature}_subdiv{subdivision}.npy", allow_pickle = True)
    elif dataset == "rwcpop":
        return np.load(f"{paths.path_data_persisted}/rwcpop/self_similarity_matrices/ssm_{dataset}_song{song_name}_barwiseTF_{similarity_type}_{feature}_subdiv{subdivision}.npy", allow_pickle = True)
    else:
        raise err.InvalidArgumentValueException("Invalid dataset name, should be either 'salami' or 'rwcpop'.")

def load_beatwise_tf_ssm(dataset, song_name, feature = "log_mel_grill", subdivision = 96, similarity_type = "cosine", train = False):
    if dataset == "salami":
        if train:
            subset = "train"
        else:
            subset = "test"
        return np.load(f"{paths.path_data_persisted}/salami/self_similarity_matrices/{subset}/ssm_{dataset}_song{song_name}_beatwiseTF_{similarity_type}_{feature}_subdiv{subdivision}.npy", allow_pickle = True)
    elif dataset == "rwcpop":
        return np.load(f"{paths.path_data_persisted}/rwcpop/self_similarity_matrices/ssm_{dataset}_song{song_name}_beatwiseTF_{similarity_type}_{feature}_subdiv{subdivision}.npy", allow_pickle = True)
    else:
        raise err.InvalidArgumentValueException("Invalid dataset name, should be either 'salami' or 'rwcpop'.")


def load_beat_sync_ssm_foote(dataset, song_name):
    """
    Load the files for the beat-sync SSM of this song, computed with the Foote particular hyperparameters (as implemented in the MSAF toolbox).
    """
    ssm = np.load(f"{paths.path_data_persisted}/{dataset}/foote_experiments/self_similarity_matrices/ssm_{dataset}_song{song_name}_beatsync.npy", allow_pickle = True)
    beat_sync_times, duration = np.load(f"{paths.path_data_persisted}/{dataset}/foote_experiments/sync_times/synctimes_{dataset}_song{song_name}_beatsync.npy", allow_pickle = True)
    return ssm, beat_sync_times, duration

def load_bar_sync_ssm_foote(dataset, song_name):
    """
    Load the files for the bar-sync SSM of this song, computed with the Foote particular hyperparameters (as implemented in the MSAF toolbox).
    """
    ssm = np.load(f"{paths.path_data_persisted}/{dataset}/foote_experiments/self_similarity_matrices/ssm_{dataset}_song{song_name}_barsync.npy", allow_pickle = True)
    bar_sync_times, duration = np.load(f"{paths.path_data_persisted}/{dataset}/foote_experiments/sync_times/synctimes_{dataset}_song{song_name}_barsync.npy", allow_pickle = True)
    return ssm, bar_sync_times, duration

def load_barwise_tf_ssm_foote(dataset, song_name, subdivision = 96):
    """
    Load the barwiseTF SSM for this song, which was persisted after a first computation, or compute it if it wasn't found.
    """
    return np.load(f"{paths.path_data_persisted}/{dataset}/foote_experiments/self_similarity_matrices/ssm_{dataset}_song{song_name}_barwiseTF_subdiv{subdivision}.npy", allow_pickle = True)

# %% Annotation loaders
def load_mirex10_annot_song_RWC(song_number):
    """
    Loads the MIREX10 annotations for this song.
    """
    annotations_mirex = f"{paths.path_annotation_rwcpop}/MIREX10"
    annot_path_mirex = f"{annotations_mirex}/{dm.get_annotation_name_from_song(song_number, 'MIREX10')}"
    annotations = dm.get_segmentation_from_txt(annot_path_mirex, "MIREX10")
    references_segments = np.array(annotations)[:,0:2]
    return references_segments

def load_RWC_dataset(music_folder_path, annotations_type = "MIREX10"):
    """
    Load the data on the RWC dataset, ie path of songs and annotations.
    The annotations can be either AIST or MIREX 10.

    Parameters
    ----------
    music_folder_path : String
        Path of the folder to parse.
    annotations_type : "AIST" [1] or "MIREX10" [2]
        The type of annotations to load (both have a specific behavior and formatting)
        The default is "MIREX10"

    Raises
    ------
    NotImplementedError
        If the format is not taken in account.

    Returns
    -------
    numpy array
        list of list of paths, each sublist being of the form [song, annotations, downbeat(if specified)].
        
    References
    ----------
    [1] Goto, M. (2006, October). AIST Annotation for the RWC Music Database. In ISMIR (pp. 359-360).
    
    [2] Bimbot, F., Sargent, G., Deruty, E., Guichaoua, C., & Vincent, E. (2014, January). 
    Semiotic description of music structure: An introduction to the Quaero/Metiss structural annotations.

    """
    # Load dataset paths at the format "song, annotations, downbeats"
    paths = []
    for file in os.listdir(music_folder_path):
        if file[-4:] == ".wav":
            file_number = "{:03d}".format(int(file[:-4]))
            ann = dm.get_annotation_name_from_song(file_number, annotations_type)
            paths.append([file, ann])
    return np.array(paths)





# %% tensor spectorgrams (not so much used, may be bugged)
def load_or_save_tensor_spectrogram(dataset, song_path, feature, hop_length, subdivision_bars, fmin = 98, n_fft = 2048, n_mfcc = 20):
    """
    Loads the BTF (bars - time - frequency) tensor for this song, which was persisted after a first computation, or compute it if it wasn't found.

    You should prefer load_or_save_spectrogram, as it allows more possibility about tensor folding, except if you're short in space on your disk.

    Parameters
    ----------
    dataset : string
        Name of the dataset. Only "rwcpop" and "salami" for now.
    song_path : string
        The path of the signal of the song.
    feature : string
        Feature of the spectrogram, part of the identifier of the spectrogram.
    hop_length : integer
        hop_length of the spectrogram, part of the identifier of the spectrogram.
    fmin : integer
        Minimal frequence for the spectrogram, part of the identifier of the spectrogram.
        The default is 98.
    n_fft and n_mfcc : integers, optional
        Both arguments are used respectively for the stft and for the mfcc computation, and are used to 

    Returns
    -------
    numpy array
        The tensor spectrogram of this song.

    """
    song_name = song_path.split("/")[-1].replace(".wav","").replace(".mp3","")
    try:
        tensor_barwise = np.load(f"{paths.path_data_persisted}/{dataset}/tensor_barwise_ae/{song_name}_{feature}_hop{hop_length}_subdiv{subdivision_bars}.npy", allow_pickle = True)
        return tensor_barwise
    except FileNotFoundError:
        the_signal, sampling_rate = librosa.load(song_path, sr=44100)
        raise NotImplementedError("Do not compute please")
        if "stft" in feature and "nfft" in feature:
            if "nfft" not in feature: 
                spectrogram = signal_to_spectrogram.get_spectrogram(the_signal, 44100, "stft", hop_length, n_fft = n_fft)
            else:              
                n_fft_arg = int(feature.split("nfft")[1])
                spectrogram = signal_to_spectrogram.get_spectrogram(the_signal, 44100, "stft", hop_length, n_fft = n_fft_arg)
        elif "mfcc" in feature:
            if "nmfcc" not in feature:
                spectrogram = signal_to_spectrogram.get_spectrogram(the_signal, 44100, "mfcc", hop_length, n_mfcc = n_mfcc)
            else:
                n_mfcc_arg = int(feature.split("nmfcc")[1])
                spectrogram = signal_to_spectrogram.get_spectrogram(the_signal, 44100, "mfcc", hop_length, n_mfcc = n_mfcc_arg)
        else:
            spectrogram = signal_to_spectrogram.get_spectrogram(the_signal, 44100, feature, hop_length, fmin = fmin)
            
        hop_length_seconds = hop_length/44100
        bars = load_or_save_bars(persisted_path, song_path)
        tensor_spectrogram = nn_utils.tensorize_barwise(spectrogram, bars, hop_length_seconds, subdivision_bars)
        np.save(f"{persisted_path}/tensor_barwise_ae/{song_name}_{feature}_hop{hop_length}_subdiv{subdivision_bars}", tensor_spectrogram)
        return tensor_spectrogram

def load_or_save_pcp_msaf(dataset, song_path, hop_length):
    """
    """
    song_name = song_path.split("/")[-1].replace(".wav","").replace(".mp3","")
    try:
        pcp, frame_times, duration = np.load(f"{paths.path_data_persisted}/{dataset}/pcp_msaf/{song_name}_hop{hop_length}.npy", allow_pickle = True)
    except FileNotFoundError:
        signal, sr = librosa.load(song_path, sr=44100)
        duration = len(signal) / float(sr)
        pcp, frame_times = signal_to_spectrogram.get_pcp_as_msaf(signal, sr, hop_length)
        np.save(f"{paths.path_data_persisted}/{dataset}/pcp_msaf/{song_name}_hop{hop_length}", (pcp, frame_times, duration))

    return pcp, frame_times, duration



# -*- coding: utf-8 -*-
"""
Created on Mon Mar 14 16:40:07 2022

@author: amarmore

Novelty cost, adapted from the work of Foote, see [1].
The kernel is of binary values : 1 and -1.
This code is deprecated, but could be used in comparison tests.

If interested, one should use the novelty version from the toolbox MSAF [2] instead,
whose parameters were tested and optimized.

References
----------
[1] J. Foote, "Automatic audio segmentation using a measure of audio novelty",
in: 2000 IEEE Int. Conf. Multimedia and Expo. ICME2000. Proc. Latest Advances 
in the Fast Changing World of Multimedia, vol. 1, IEEE, 2000,pp. 452–455.

[2] O. Nieto and J.P. Bello, "Systematic exploration of computational music
structure research.", in: ISMIR, 2016, pp. 547–553.
"""
# %% Using implementation from msaf
import msaf
import numpy as np
from scipy import signal
from scipy.ndimage import filters

def compute_foote_ssm(input_spectrogram, pre_filter = 12):
    spec = input_spectrogram
    if pre_filter != 0:
        median_filtering_size = pre_filter
        spec = median_filter(input_spectrogram, M=median_filtering_size)
        
    return msaf.algorithms.foote.segmenter.compute_ssm(spec)

def process_msaf_own_as(input_spectrogram = None, input_ssm = None, M_gaussian = 66, L_peaks = 64, pre_filter = 12, post_filter = 4):
        """Main process.
        Returns
        -------
        est_idxs : np.array(N)
            Estimated indeces the segment boundaries in frames.
        est_labels : np.array(N-1)
            Estimated labels for the segments.
        """
        assert input_ssm is not None or input_spectrogram is not None, "Either input_ssm or input_spectrogram should be provided"

        # Make sure that the M_gaussian is even
        if M_gaussian % 2 == 1:
            M_gaussian += 1

        # Self similarity matrix
        if input_ssm is None:
            S = compute_foote_ssm(input_spectrogram, pre_filter)
            input_shape = input_spectrogram.shape
        else:
            S = input_ssm
            input_shape = input_ssm.shape

        # Compute gaussian kernel
        G = msaf.algorithms.foote.segmenter.compute_gaussian_krnl(M_gaussian)
        #plt.imshow(S, interpolation="nearest", aspect="auto"); plt.show()

        # Compute the novelty curve
        nc = msaf.algorithms.foote.segmenter.compute_nc(S, G)

        # Find peaks in the novelty curve
        sigma = post_filter
        est_idxs = pick_peaks(nc, sigma = sigma, L=L_peaks)

        # Add first and last frames
        est_idxs = np.concatenate(([0], est_idxs, [input_shape[0] - 1]))

        # Empty labels
        est_labels = np.ones(len(est_idxs) - 1) * -1

        # Post process estimations
        est_idxs, est_labels = postprocess_msaf(est_idxs, est_labels)

        return est_idxs, est_labels
        
def median_filter(X, M=8):
    """Median filter along the first axis of the feature matrix X."""
    for i in range(X.shape[1]):
        X[:, i] = filters.median_filter(X[:, i], size=M)
    return X

def pick_peaks(nc, sigma = 4, L=16):
    """Obtain peaks from a novelty curve using an adaptive threshold."""
    offset = nc.mean() / 20.

    if sigma != 0:
        nc = filters.gaussian_filter1d(nc, sigma=sigma)  # Smooth out nc

    th = filters.median_filter(nc, size=L) + offset
    #th = filters.gaussian_filter(nc, sigma=L/2., mode="nearest") + offset

    peaks = []
    for i in range(1, nc.shape[0] - 1):
        # is it a peak?
        if nc[i - 1] < nc[i] and nc[i] > nc[i + 1]:
            # is it above the threshold?
            if nc[i] > th[i]:
                peaks.append(i)
    #plt.plot(nc)
    #plt.plot(th)
    #for peak in peaks:
        #plt.axvline(peak)
    #plt.show()

    return peaks
        
def postprocess_msaf(est_idxs, est_labels):
        """Post processes the estimations from the algorithm, removing empty
        segments and making sure the lenghts of the boundaries and labels
        match."""
        # Make sure we are using the previously input bounds, if any
        # if self.in_bound_idxs is not None:
            # F = self._preprocess()
            # est_labels = msaf.utils.synchronize_labels(self.in_bound_idxs, est_idxs,
                                              # est_labels, F.shape[0])
            # est_idxs = self.in_bound_idxs

        # Remove empty segments if needed
        est_idxs, est_labels = msaf.utils.remove_empty_segments(est_idxs, est_labels)

        assert len(est_idxs) - 1 == len(est_labels), "Number of boundaries " \
            "(%d) and number of labels(%d) don't match" % (len(est_idxs),
                                                           len(est_labels))

        # Make sure the indeces are integers
        est_idxs = np.asarray(est_idxs, dtype=int)

        return est_idxs, est_labels

def my_process_segmentation_level(est_idxs, est_labels, N, frame_times, dur):
    """Processes a level of segmentation, and converts it into times.

    Parameters
    ----------
    est_idxs: np.array
        Estimated boundaries in frame indeces.
    est_labels: np.array
        Estimated labels.
    N: int
        Number of frames in the whole track.
    frame_times: np.array
        Time stamp for each frame.
    dur: float
        Duration of the audio track.

    Returns
    -------
    est_times: np.array
        Estimated segment boundaries in seconds.
    est_labels: np.array
        Estimated labels for each segment.
    """
    assert est_idxs[0] == 0 and est_idxs[-1] == N - 1
    assert len(est_idxs) - 1 == len(est_labels)

    # Add silences, if needed
    est_times = np.concatenate(([0], frame_times[est_idxs], [dur]))
    silence_label = np.max(est_labels) + 1
    est_labels = np.concatenate(([silence_label], est_labels, [silence_label]))

    # Remove empty segments if needed
    est_times, est_labels = msaf.utils.remove_empty_segments(est_times, est_labels)

    # I've added these lines because bar estimates may be less precise than frame and beats.
    if est_times[-1] < dur: # Fixing a bug, due to empty segment: [dur, dur]
        est_times.append(dur)
    if est_times[-1] > dur: # Fixing a bug, due to inaccuracies (rounding errors mainly)
        if est_times[-1] - dur < 1:
            est_times[-1] = dur
        else:
            #print(f"Ajout Axel: est_times[-1] = {est_times[-1]}, dur = {dur}, ce qui n'est pas normal. Voir si bug arrive souvent.")
            est_times[-1] = dur
    
    # Make sure that the first and last times are 0 and duration, respectively
    assert np.allclose([est_times[0]], [0]) and np.allclose([est_times[-1]], [dur])

    return est_times, est_labels


# %% Novelty computation
def novelty_cost(cropped_autosimilarity):
    """
    Novelty measure on this part of the autosimilarity matrix.
    The size of the kernel will be the size of the parameter matrix.

    Parameters
    ----------
    cropped_autosimilarity : list of list of floats or numpy array (matrix representation)
        The part of the autosimilarity which novelty measure is to compute.

    Raises
    ------
    NotImplementedError
        If the size of the autosimilarity is odd (novlety kernel can't fit this matrix).

    Returns
    -------
    float
        The novelty measure.

    """
    # Kernel is of the size of cropped_autosimilarity
    if len(cropped_autosimilarity) == 0:
        return 0
    
    if len(cropped_autosimilarity) % 2 == 1:
        raise NotImplementedError("The novelty computation is not implemented when the kernel is of odd size.") from None
        #return (novelty_cost(cropped_autosimilarity[:-1, :-1]) + novelty_cost(cropped_autosimilarity[1:, 1:])) / 2
    
    kernel_size = int(len(cropped_autosimilarity) / 2)
    kernel = np.kron(np.array([[1,-1], [-1, 1]]), np.ones((kernel_size, kernel_size)))
    return np.mean(kernel*cropped_autosimilarity)

def novelty_computation(autosimilarity_array, kernel_size):
    """
    Computes the novelty measure on the entire autosimilarity matrix, with a defined and fixed kernel size.

    Parameters
    ----------
    autosimilarity_array : list of list of floats or numpy array (matrix representation)
        The autosimilarity matrix.

    kernel_size : integer
        The size of the kernel.

    Raises
    ------
    NotImplementedError
        If the kernel size is odd, can't compute the novelty measure.

    Returns
    -------
    cost : list of float
        List of novelty measures, at each bar of the autosimilarity.

    """
    if kernel_size % 2 == 1:
        raise NotImplementedError("The novelty computation is not implemented when the kernel is of odd size.") from None
    cost = np.zeros(len(autosimilarity_array))
    half_kernel = int(kernel_size / 2)
    for i in range(half_kernel, len(autosimilarity_array) - half_kernel):
        cost[i] = novelty_cost(autosimilarity_array[i - half_kernel:i + half_kernel,i - half_kernel:i + half_kernel])
    return cost

# %% Post-processing of the novelty values
##################### Sandbox ##################
def peak_picking(tab, window_size = 1):
    """
    Returns the indexes of peaks of values in the given list of values.
    A value is considered "peak" if it's a local maximum,
    and if all values in the window (defined by 'window_size') before and after 
    are strictly monotonous.    
    Used for peak picking in the novelty measure.

    Parameters
    ----------
    tab : list of float
        The list of values to study.
    window_size : boolean, optional
        Size of the window around a possible peak to be considered "peak",
        ie number of consecutive values where the values should increase (before) and (decrease) after.
        The default is 1.

    Returns
    -------
    to_return : list of integers
        The indexes where values are peaking.

    """
    to_return = []
    for current_idx in range(window_size, len(tab) - window_size):
        if is_increasing(tab[current_idx - window_size:current_idx + 1]) and is_increasing(tab[current_idx:current_idx + window_size + 1][::-1]):
            to_return.append(current_idx)
    return to_return

def valley_picking(tab, window_size = 1):
    """
    Returns the indexes of valleys of values in the desired list of values.
    A value is considered "valley" if it's a local minimum,
    and if all values in the window (defined by 'window_size') before and after 
    are strictly monotonous.
    Used for peak picking in the novelty measure.

    Parameters
    ----------
    tab : list of float
        The list of values to study.
    window_size : boolean, optional
        Size of the window around a possible valley to be considered "valley",
        ie number of consecutive values where the values should decrease (before) and increase (after).
        The default is 1.

    Returns
    -------
    to_return : list of integers
        The indexes where values are valleys.

    """
    to_return = []
    for current_idx in range(window_size, len(tab) - window_size):
        if is_increasing(tab[current_idx - window_size:current_idx + 1][::-1]) and is_increasing(tab[current_idx:current_idx + window_size + 1]):
            to_return.append(current_idx)
    return to_return

def is_increasing(tab):
    """
    Tests if the tab values are increasing.
    Used for peak picking in the novelty measure.

    Parameters
    ----------
    tab : list of float
        The values.

    Returns
    -------
    boolean
        Whether the values are increasing or not.

    """
    if len(tab) <= 1 or len(np.unique(tab)) == 1:
        return False
    for idx in range(len(tab) - 1):
        if tab[idx] > tab[idx+1]:
            return False
    return True

def decreasing_peaks(data):
    """
    Returns the peaks indexes of a list of values in their decreasing order of values.
    Used for peak picking in the novelty measure.

    Parameters
    ----------
    data : list of float
        The values.

    Returns
    -------
    list of integers
        The indexes of the peaks, sorted in their decreasing order of values.

    """
    peaks_and_value = []
    for idx in peak_picking(data, window_size = 1):
        peaks_and_value.append((idx, data[idx]))
    return sorted(peaks_and_value, key=lambda x:x[1], reverse = True)

def select_highest_peaks_thresholded_indexes(data, percentage = 0.33):
    """
    Returns the peaks higher than a percentage of the maximal peak from a list of values.
    Used for peak picking in the novelty measure.
    
    Parameters
    ----------
    data : list of floats
        The values.
    percentage : float, optional
        The percentage of the maximal value for a peak to be valid.
        The default is 0.33.

    Returns
    -------
    list of integers
        Indexes of the valid peaks.

    """
    peaks = np.array(decreasing_peaks(data))
    max_peak = peaks[0,1]
    for idx, peak in enumerate(peaks):
        if peak[1] < percentage * max_peak:
            return [int(i) for i in sorted(peaks[:idx, 0])]
    return [int(i) for i in sorted(peaks[:,0])]

def values_as_slop(value, choice_func = max):
    """
    Compute peaks of a value (typically novelty measure)
    as the difference between absolute peaks and absolute valleys.
    Function choice_func determines the way of computing this gap.
    
    Typically, max will compute peaks as the maximum gap between a peaks and its two closest valleys,
    whereas min will select the minimal gap.
    
    This returns an array containing zeroes where there is no peak in absoluite value,
    and this new value as a gap computation where there was peaks before.

    Parameters
    ----------
    value : array of float
        The absolute value of the measure.
    choice_func : function name, optional
        Type of the function selecting the difference between peaks and valleys.
        Classical values are "max" for selecting the maximum gap between the peak and both its closest valleys,
        "min" for the minimum of both gaps, and "mean" (called as_seg.mean) for the mean of both gaps.
        The default is max.

    Returns
    -------
    peak_valley_slop : array of floats
        The new values of peaks as gaps, and 0 everywhere else.

    """
    peaks = peak_picking(value, window_size = 1)
    valleys = valley_picking(value, window_size = 1)
    peak_valley_slop = np.zeros(len(value))
    for peak in peaks:
        i = 0
        while i < len(valleys) and valleys[i] < peak:
            i+=1
        if i == 0:
            left_valley = 0
            right_valley = valleys[i]
        elif i == len(valleys):
            left_valley = valleys[i - 1]
            right_valley = 0
        else:
            left_valley = valleys[i - 1]
            right_valley = valleys[i]
        chosen_valley_value = choice_func(value[left_valley], value[right_valley])
        peak_valley_slop[peak] = value[peak] - chosen_valley_value
    return peak_valley_slop
    
def mean(val_a, val_b):
    """
    A function returning the mean of both values.
    This function is redeveloped so as to be called as choice_func in the function "values_as_slop()" (see above) in external projects.

    Parameters
    ----------
    val_a : float
        First value.
    val_b : float
        Second value.

    Returns
    -------
    float: mean of both values.

    """
    return (val_a + val_b) / 2


# -*- coding: utf-8 -*-
"""
Created on Fri Feb 21 15:09:10 2020

@author: amarmore

A file which contains all code regarding conversion of data, or information extraction from it
(typically getting the bars, converting segments in frontiers, sonifying segmentation or computing its Hit-Rate score, etc).
"""

import as_seg.model.errors as err

import numpy as np
import madmom.features.downbeats as dbt
import madmom.features.beats as bt
import mir_eval
import scipy
import librosa
import math

# %% Read and treat inputs
def get_bars_from_audio(song_path):
    """
    Returns the bars of a song, directly from its audio signal.
    Encapsulates the downbeat estimator from the madmom toolbox [1].
    
    NB1: Note that the estimation implicitely assumes 3 or 4 beats per bar.
    
    NB2: Note also that this function artificially adds bars at the end of the song, so that the estimation spans the entire song length.
    May/should be debated.

    Parameters
    ----------
    song_path : String
        Path to the desired song.

    Returns
    -------
    downbeats_times : list of tuples of float
        List of the estimated bars, as (start, end) times.
        
    References
    ----------
    [1] Böck, S., Korzeniowski, F., Schlüter, J., Krebs, F., & Widmer, G. (2016, October). 
    Madmom: A new python audio and music signal processing library. 
    In Proceedings of the 24th ACM international conference on Multimedia (pp. 1174-1178).

    """
    act = dbt.RNNDownBeatProcessor()(song_path)
    proc = dbt.DBNDownBeatTrackingProcessor(beats_per_bar=[3,4], fps=100)
    song_beats = proc(act)
    downbeats_times = [song_beats[0][0]]
    
    for beat in song_beats[1:]: # The first beat is already added
        if beat[1] == 1: # If the beat is a downbeat
            downbeats_times.append(beat[0])
            
    song_length = act.shape[0]/100 # Total length of the song
    downbeats_times.append(song_length) # adding the last downbeat
    return frontiers_to_segments(downbeats_times)
    
def get_beats_from_audio_madmom(song_path):
    """
    Uses madmom to estimate the beats of a song, from its audio signal.
    """
    act = bt.TCNBeatProcessor()(song_path)
    proc = bt.BeatTrackingProcessor(fps=100)
    song_beats = proc(act)
    
    # beats_times = []    
    # if song_beats[0][1] != 1: # Adding a first downbeat at the start of the song
        # beats_times.append(0.1)
    # for beat in song_beats:
        # if beat[1] == 1: # If the beat is a downbeat
            # downbeats_times.append(beat[0])
            
    return frontiers_to_segments(list(song_beats))

def get_beats_from_audio_msaf(signal, sr, hop_length):
    """
    Uses MSAF to estimate the beats of a song, from its audio signal.
    """
    _, audio_percussive = librosa.effects.hpss(signal)
    
    # Compute beats
    _, beat_frames = librosa.beat.beat_track(y=audio_percussive, sr=sr, hop_length=hop_length)

    # To times
    beat_times = librosa.frames_to_time(beat_frames, sr=sr,hop_length=hop_length)

    # TODO: Is this really necessary?
    if len(beat_times) > 0 and beat_times[0] == 0:
        beat_times = beat_times[1:]
        beat_frames = beat_frames[1:]
        
    return beat_times, beat_frames
    
# %% Read and treat inputs
def get_segmentation_from_txt(path, annotations_type):
    """
    Reads the segmentation annotations, and returns it in a list of tuples (start, end, index as a number)
    This function has been developped for AIST and MIREX10 annotations, adapted for these types of annotations.
    It will not work with another set of annotation.

    Parameters
    ----------
    path : String
        The path to the annotation.
    annotations_type : "AIST" [1] or "MIREX10" [2]
        The type of annotations to load (both have a specific behavior and formatting)
        
    Raises
    ------
    err.InvalidArgumentValueException
        If the type of annotations is neither AIST or MIREX10

    Returns
    -------
    segments : list of tuples (float, float, integer)
        The segmentation, formatted in a list of tuples, and with labels as numbers (easier to interpret computationnally).

    References
    ----------
    [1] Goto, M. (2006, October). AIST Annotation for the RWC Music Database. In ISMIR (pp. 359-360).
    
    [2] Bimbot, F., Sargent, G., Deruty, E., Guichaoua, C., & Vincent, E. (2014, January). 
    Semiotic description of music structure: An introduction to the Quaero/Metiss structural annotations.

    """
    file_seg = open(path)
    segments = []
    labels = []
    for part in file_seg.readlines():
        tupl = part.split("\t")
        if tupl[2] not in labels: # If label wasn't already found in this annotation
            idx = len(labels)
            labels.append(tupl[2])
        else: # If this label was found for another segment
            idx = labels.index(tupl[2])
        if annotations_type == "AIST":
            segments.append(((int(tupl[0]) / 100), (int(tupl[1]) / 100), idx))
        elif annotations_type == "MIREX10":
            segments.append((round(float(tupl[0]), 3), round(float(tupl[1]), 3), idx))
        else:
            raise err.InvalidArgumentValueException("Annotations type not understood")
    return segments

def get_annotation_name_from_song(song_number, annotations_type):
    """
    Returns the name of the annotation of this song according to the desired annotation type
    
    Specificly designed for RWC Pop dataset, shouldn't be used otherwise.
    For now are available:
        - AIST annotations [1]
        - MIREX 10 annotations [2]
    
    Parameters
    ----------
    song_number : integer or string
        The number of the song (which is its name).
    annotations_type : string
        The desired type of annotation.

    Raises
    ------
    InvalidArgumentValueException
        If the annotatipn type is not implemented.

    Returns
    -------
    string
        The name of the file containing the annotation.
        
    References
    ----------
    [1] Goto, M. (2006, October). AIST Annotation for the RWC Music Database. In ISMIR (pp. 359-360).
    
    [2] Bimbot, F., Sargent, G., Deruty, E., Guichaoua, C., & Vincent, E. (2014, January). 
    Semiotic description of music structure: An introduction to the Quaero/Metiss structural annotations.

    """
    if annotations_type == "MIREX10":
        return "RM-P{:03d}.BLOCKS.lab".format(int(song_number))
    elif annotations_type == "AIST":
        return "RM-P{:03d}.CHORUS.TXT".format(int(song_number))
    else:
        raise err.InvalidArgumentValueException("Annotations type not understood")

# %% Conversion of data (time/frame/beat and segment/frontiers)
def frontiers_from_time_to_frame_idx(seq, hop_length_seconds):
    """
    Converts a sequence of frontiers in time to their values in frame indexes.

    Parameters
    ----------
    seq : list of float/times
        The list of times to convert.
    hop_length_seconds : float
        hop_length (time between two consecutive frames), in seconds.

    Returns
    -------
    list of integers
        The sequence, as a list, in frame indexes.
    """
    
    return [int(round(frontier/hop_length_seconds)) for frontier in seq]

def segments_from_time_to_frame_idx(segments, hop_length_seconds):
    """
    Converts a sequence of segments (start, end) in time to their values in frame indexes.

    Parameters
    ----------
    segements : list of tuple
        The list of segments, as tuple (start, end), to convert.
    hop_length_seconds : float
        hop_length (time between two consecutive frames), in seconds.

    Returns
    -------
    list of integers
        The sequence, as a list, in frame indexes.
    """
    to_return = []
    for segment in segments:
        bar_in_frames = [int(round(segment[0]/hop_length_seconds)), int(round(segment[1]/hop_length_seconds))]
        if bar_in_frames[0] != bar_in_frames[1]:
            to_return.append(bar_in_frames)
    return to_return
    
def segments_from_time_to_bar(segments, bars):
    """
    Converts the segments in time to segments in bar indexes.
    The selected bar is the one which end is the closest from the frontier.

    Parameters
    ----------
    segments : list of tuples
        The list of segments, in time.
    bars : list of tuple of floats
        The bars, as (start time, end time) tuples.

    Returns
    -------
    list of tuples of integers
        List of time instances (start, end) converted in bar indexes.

    """
    frontiers = segments_to_frontiers(segments)
    return np.array(frontiers_to_segments(frontiers_from_time_to_bar(frontiers, bars)))

def frontiers_from_time_to_bar(seq, bars):
    """
    Converts the frontiers in time to a bar index.
    The selected bar is the one which end is the closest from the frontier.

    Parameters
    ----------
    seq : list of float
        The list of frontiers, in time.
    bars : list of tuple of floats
        The bars, as (start time, end time) tuples.

    Returns
    -------
    seq_barwise : list of integers
        List of times converted in bar indexes.

    """
    seq_barwise = []
    for frontier in seq:
        for idx, bar in enumerate(bars):
            if frontier >= bar[0] and frontier < bar[1]:
                if bar[1] - frontier < frontier - bar[0]:
                    seq_barwise.append(idx)
                else:
                    if idx == 0:
                        seq_barwise.append(idx)
                        #print("The current frontier {} is labelled in the start silence ({},{}), which is incorrect.".format(frontier, bar[0], bar[1]))
                    else:
                        seq_barwise.append(idx - 1)
                break
    return seq_barwise

def frontiers_from_bar_to_time(seq, bars):
    """
    Converts the frontiers (or a sequence of integers) from bar indexes to absolute times of the bars.
    The frontier is considered as the end of the bar.

    Parameters
    ----------
    seq : list of integers
        The frontiers, in bar indexes.
    bars : list of tuple of floats
        The bars, as (start time, end time) tuples.

    Returns
    -------
    to_return : list of float
        The frontiers, converted in time (from bar indexes).

    """
    to_return = []
    for frontier in seq:
        bar_frontier = bars[frontier][1]
        if bar_frontier not in to_return:
            to_return.append(bar_frontier)
    return to_return

def segments_from_bar_to_time(segments, bars):
    """
    Converts segments from bar indexes to time.

    Parameters
    ----------
    segments : list of tuple of integers
        The indexes of the bars defining the segments (start, end).
    bars : list of tuple of float
        Bars, as tuples (start, end), in time.

    Returns
    -------
    numpy array
        Segments, in time.

    """
    to_return = []
    for start, end in segments:
        if end >= len(bars):
            to_return.append([bars[start][1], bars[-1][1]])
        else:
            to_return.append([bars[start][1], bars[end][1]])
    return np.array(to_return)
    
def frontiers_to_segments(frontiers):
    """
    Computes a list of segments starting from the frontiers between them.

    Parameters
    ----------
    frontiers : list of floats
        The list of frontiers.

    Returns
    -------
    to_return : list of tuples of floats
        The segments, as tuples (start, end).

    """
    to_return = []
    while 0 in frontiers:
        frontiers.remove(0)
    to_return.append((0,frontiers[0]))
    for idx in range(len(frontiers) - 1):
        if frontiers[idx] != frontiers[idx + 1]:
            to_return.append((frontiers[idx], frontiers[idx + 1]))
    return to_return

def segments_to_frontiers(segments):
    """
    Computes a list of frontiers from the segments.

    Parameters
    ----------
    segments : list of tuples of floats
        The segments, as tuples.

    Returns
    -------
    list
        Frontiers between segments.

    """
    return [i[1] for i in segments]

def align_segments_on_bars(segments, bars):
    """
    Aligns the estimated segments to the closest bars (in time).
    The idea is that segments generally start and end on downbeats,
    and that realigning the estimation could improve perfomance for low tolerances scores.
    Generally used for comparison with techniques which don't align their segmentation on bars.

    Parameters
    ----------
    segments : list of tuple of float
        Time of the estimated segments, as (start, end).
    bars : list of tuple of float
        The bars of the signal.

    Returns
    -------
    list of tuple of floats
        Segments, realigned on bars.

    """
    frontiers = segments_to_frontiers(segments)
    return frontiers_to_segments(align_frontiers_on_bars(frontiers, bars))

def align_frontiers_on_bars(frontiers, bars):
    """
    Aligns the frontiers of segments to the closest bars (in time).
    The idea is that frontiers generally occurs on downbeats,
    and that realigning the estimation could improve perfomance for low tolerances scores.
    Generally used for comparison with techniques which don't align their segmentation on bars.

    Parameters
    ----------
    frontiers : list of float
        Time of the estimated frontiers.
    bars : list of tuple of float
        The bars of the signal.

    Returns
    -------
    frontiers_on_bars : list of floats
        Frontiers, realigned on bars.

    """
    frontiers_on_bars = []
    i = 1
    for frontier in frontiers:
        while i < len(bars) - 1 and bars[i][1] < frontier:
            i+=1
        if i == len(bars) - 1:
            frontiers_on_bars.append(frontier)
        else:
            if bars[i][1] - frontier < frontier - bars[i][0]:
                frontiers_on_bars.append(bars[i][1])
            else:
                frontiers_on_bars.append(bars[i][0])
    return frontiers_on_bars

def get_median_hop(bars, subdivision = 96, sampling_rate = 44100):
    """
    Returns the median hop length in the song, used for audio reconstruction.
    The rationale is that all bars are sampled with 'subdivision' number of frames, 
    but they can be of different lengths in absolute time.
    Hence, the time gap between two consecutive frames (the hop length) can differ between bars.
    For reconstruction, we use the median hop length among all bars.

    Parameters
    ----------
    bars : list of tuples of float
        The bars, as (start time, end time) tuples.
    subdivision : integer, optional
        The number of subdivision of the bar to be contained in each slice of the tensor.
        The default is 96.
    sampling_rate : integer, optional
        The sampling rate of the signal, in Hz.
        The default is 44100.

    Returns
    -------
    integer
        The median hop length in these bars.

    """
    hops = []
    for bar_idx in range(1, len(bars)):
        len_sig = bars[bar_idx][1] - bars[bar_idx][0]
        hop = int(len_sig/subdivision * sampling_rate)
        hops.append(hop)
    return int(np.median(hops)) # Generally used for audio reconstruction

# %% Sonification of the segmentation
def sonify_frontiers_path(audio_file_path, frontiers_in_seconds, output_path):
    """
    Takes the path of the song and frontiers, and write a song with the frontiers sonified ("bip" in the song).
    Function inspired from MSAF.

    Parameters
    ----------
    audio_file_path: String
        The path to the song, (as signal).
    frontiers_in_seconds: list of floats
        The frontiers, in time/seconds.
    output_path: String
        The path where to write the song with sonified frontiers.

    Returns
    -------
    Nothing, but writes a song at output_path

    """
    the_signal, sampling_rate = librosa.load(audio_file_path, sr=None)
    sonify_frontiers_song(the_signal, sampling_rate, frontiers_in_seconds, output_path)

def sonify_frontiers_song(song_signal, sampling_rate, frontiers_in_seconds, output_path):
    """
    Takes a song as a signal, and add the frontiers to this signal.
    It then writes it as a file.
    Function inspired from MSAF.

    Parameters
    ----------
    song_signal : numpy array
        The song as a signal.
    sampling_rate : integer
        The sampling rate of the signal, in Hz.
    frontiers_in_seconds: list of floats
        The frontiers, in time/seconds.
    output_path: String
        The path where to write the song with sonified frontiers.

    Returns
    -------
    Nothing, but writes a song at the output_path.

    """
    frontiers_signal = mir_eval.sonify.clicks(frontiers_in_seconds, sampling_rate)
    
    singal_with_frontiers = np.zeros(max(len(song_signal[:,0]), len(frontiers_signal)))
    
    singal_with_frontiers[:len(song_signal[:,0])] = song_signal[:,0]
    singal_with_frontiers[:len(frontiers_signal)] += frontiers_signal
    
    scipy.io.wavfile.write(output_path, sampling_rate, singal_with_frontiers)
    
# %% Score calculation encapsulation
def compute_score_from_frontiers_in_bar(reference, frontiers_in_bar, bars, window_length = 0.5):
    """
    Computes precision, recall and f measure from estimated frontiers (in bar indexes) and the reference (in seconds).
    Scores are computed from the mir_eval toolbox.

    Parameters
    ----------
    reference : list of tuples
        The reference annotations, as a list of tuples (start, end), in seconds.
    frontiers : list of integers
        The frontiers between segments, in bar indexes.
    bars : list of tuples
        The bars of the song.
    window_length : float, optional
        The window size for the score (tolerance for the frontier to be validated).
        The default is 0.5.

    Returns
    -------
    precision: float \in [0,1]
        Precision of these frontiers,
        ie the proportion of accurately found frontiers among all found frontiers.
    recall: float \in [0,1]
        Recall of these frontiers,
        ie the proportion of accurately found frontiers among all accurate frontiers.
    f_measure: float \in [0,1]
        F measure of these frontiers,
        ie the geometric mean of both precedent scores.
        
    """
    try:
        np.array(bars).shape[1]
    except:
        raise err.OutdatedBehaviorException("Bars is still a list of downbeats, which is an old beavior, and shouldn't happen anymore. To track and to fix.")
    frontiers_in_time = frontiers_from_bar_to_time(frontiers_in_bar, bars)
    return compute_score_of_segmentation(reference, frontiers_to_segments(frontiers_in_time), window_length = window_length)

def compute_score_of_segmentation(reference, segments_in_time, window_length = 0.5, trim = False):
    """
    Computes precision, recall and f measure from estimated segments and the reference, both in seconds.    
    Scores are computed from the mir_eval toolbox.

    Parameters
    ----------
    reference : list of tuples
        The reference annotations, as a list of tuples (start, end), in seconds.
    segments_in_time : list of tuples
        The segments, in seconds, as tuples (start, end).
    window_length : float, optional
        The window size for the score (tolerance for the frontier to be validated).
        The default is 0.5.
    trim : boolean, optional
        Whether (True) or not (False) the first and last boundaries should be "trimmed", 
        i.e. discarded in the computation of accurate boundaries.
        Default is False, meaning that they are kept by default.
        This may artifically increase the metrics, but is the standard in the benchmarks.

    Returns
    -------
    precision: float \in [0,1]
        Precision of these frontiers,
        ie the proportion of accurately found frontiers among all found frontiers.
    recall: float \in [0,1]
        Recall of these frontiers,
        ie the proportion of accurately found frontiers among all accurate frontiers.
    f_measure: float \in [0,1]
        F measure of these frontiers,
        ie the geometric mean of both precedent scores.

    """
    ref_intervals, useless = mir_eval.util.adjust_intervals(reference,t_min=0)
    est_intervals, useless = mir_eval.util.adjust_intervals(np.array(segments_in_time), t_min=0, t_max=ref_intervals[-1, 1])
    try:
        return mir_eval.segment.detection(ref_intervals, est_intervals, window = window_length, trim = trim)
    except ValueError:
        cleaned_intervals = []
        #print("A segment is (probably) composed of the same start and end. Can happen with time -> bar -> time conversion, but should'nt happen for data originally segmented in bars.")
        for idx in range(len(est_intervals)):
            if est_intervals[idx][0] != est_intervals[idx][1]:
                cleaned_intervals.append(est_intervals[idx])
        return mir_eval.segment.detection(ref_intervals, np.array(cleaned_intervals), window = window_length, trim = trim)

def compute_median_deviation_of_segmentation(reference, segments_in_time):
    """
    TODO

    Parameters
    ----------
    reference : list of tuples
        The reference annotations, as a list of tuples (start, end), in seconds.
    segments_in_time : list of tuples
        The segments, in seconds, as tuples (start, end).

    Returns
    -------
    TODO
    r_to_e then e_to_r

    """
    ref_intervals, useless = mir_eval.util.adjust_intervals(reference,t_min=0)
    est_intervals, useless = mir_eval.util.adjust_intervals(np.array(segments_in_time), t_min=0, t_max=ref_intervals[-1, 1])
    try:
        return mir_eval.segment.deviation(ref_intervals,est_intervals)
    except ValueError:
        cleaned_intervals = []
        for idx in range(len(est_intervals)):
            if est_intervals[idx][0] != est_intervals[idx][1]:
                cleaned_intervals.append(est_intervals[idx])
        return mir_eval.segment.deviation(ref_intervals,est_intervals)

def compute_rates_of_segmentation(reference, segments_in_time, window_length = 0.5):
    """
    Computes True Positives, False Positives and False Negatives from estimated segments and the reference, both in seconds.    
    Scores are computed from the mir_eval toolbox. (In fact, the code is extracted from mir_eval.segment.detection)

    Parameters
    ----------
    reference : list of tuples
        The reference annotations, as a list of tuples (start, end), in seconds.
    segments_in_time : list of tuples
        The segments, in seconds, as tuples (start, end).
    window_length : float, optional
        The window size for the score (tolerance for the frontier to be validated).
        The default is 0.5.

    Returns
    -------
    True Positives: Integer
        The number of True Positives, 
        ie the number of accurately found frontiers.
    False Positives: Integer
        The number of False Positives,
        ie the number of wrongly found frontiers (estimated frontiers which are incorrect).
    False Negative : Integer
        The number of False Negatives,
        ie the number of frontiers undetected (accurate frontiers which are not found in teh estimation).

    """  
    reference_intervals, _ = mir_eval.util.adjust_intervals(reference,t_min=0)
    estimated_intervals, _ = mir_eval.util.adjust_intervals(segments_in_time, t_min=0, t_max=reference_intervals[-1, 1])
    
    mir_eval.segment.validate_boundary(reference_intervals, estimated_intervals, False)

    # Convert intervals to boundaries
    reference_boundaries = mir_eval.util.intervals_to_boundaries(reference_intervals)
    estimated_boundaries = mir_eval.util.intervals_to_boundaries(estimated_intervals)

    # If we have no boundaries, we get no score.
    if len(reference_boundaries) == 0 or len(estimated_boundaries) == 0:
        return 0, 0, 0

    tp = len(mir_eval.util.match_events(reference_boundaries,estimated_boundaries,window_length))
    fp = len(estimated_boundaries) - tp
    fn = len(reference_boundaries) - tp
    
    return tp, fp, fn
        

# %% High level encapsulation of the computation of scores, based on segments.
## Tolerances are MIREX standards in time (0.5s and 3s), or 0 and 1 bar when barwise aligned.
def get_scores_from_segments_in_time(segments_in_time, ref_tab):
    """
    Computes the scores of the segmentation from the segments in time and the references when references may be multiple.
    """
    if type(ref_tab[0][0]) != np.ndarray: # ref_tab consist in the references, and should be nested in an array (for consistency).
        ref_tab = [ref_tab]

    res = -math.inf * np.ones((2, 3))

    prec05, rap05, f_mes05 = compute_score_of_segmentation(ref_tab[0], segments_in_time, window_length = 0.5)
    prec3, rap3, f_mes3 = compute_score_of_segmentation(ref_tab[0], segments_in_time, window_length = 3)
    res = [[round(prec05,4),round(rap05,4),round(f_mes05,4)], [round(prec3,4),round(rap3,4),round(f_mes3,4)]]
    
    if len(ref_tab) > 1:
        nd_prec05, nd_rap05, nd_f_mes05 = compute_score_of_segmentation(ref_tab[1], segments_in_time, window_length = 0.5)
        nd_prec3, nd_rap3, nd_f_mes3 = compute_score_of_segmentation(ref_tab[1], segments_in_time, window_length = 3)
        if nd_f_mes05 + nd_f_mes3 > f_mes05 + f_mes3:
            res = [[round(nd_prec05,4),round(nd_rap05,4),round(nd_f_mes05,4)], [round(nd_prec3,4),round(nd_rap3,4),round(nd_f_mes3,4)]]
    
    return res

def get_scores_in_time_from_barwise_segments(segments, bars, ref_tab):
    """
    Computes the scores of the segmentation from the segments in bar indexes and the references.
    """
    segments_in_time = segments_from_bar_to_time(segments, bars)
    return get_scores_from_segments_in_time(segments_in_time, ref_tab)
    
def get_scores_in_bars_from_barwise_segments(segments, bars, ref_tab):
    """
    Get scores from segments in bar indexes and references, with tolerance expressed in bars.
    """
    res = -math.inf * np.ones((2, 3))

    if type(ref_tab[0][0]) != np.ndarray: # ref_tab consist in the references, and should be nested in an array (for consistency between double anntoations in SALAMI and single in RWC).
        ref_tab = [ref_tab]

    ref0_in_bars = np.array(segments_from_time_to_bar(ref_tab[0], bars))
    
    prec0bar, rap0bar, f_mes0bar = compute_score_of_segmentation(ref0_in_bars, segments, window_length = 0)
    prec1bar, rap1bar, f_mes1bar = compute_score_of_segmentation(ref0_in_bars, segments, window_length = 1)
    res = [[round(prec0bar,4),round(rap0bar,4),round(f_mes0bar,4)], [round(prec1bar,4),round(rap1bar,4),round(f_mes1bar,4)]]

    if len(ref_tab) > 1:
        ref1_in_bars = np.array(segments_from_time_to_bar(ref_tab[1], bars))

        nd_prec0bar, nd_rap0bar, nd_f_mes0bar = compute_score_of_segmentation(ref1_in_bars, segments, window_length = 0)
        nd_prec1bar, nd_rap1bar, nd_f_mes1bar = compute_score_of_segmentation(ref1_in_bars, segments, window_length = 1)
        if nd_f_mes0bar + nd_f_mes1bar > f_mes0bar + f_mes1bar:
            res = [[round(nd_prec0bar,4),round(nd_rap0bar,4),round(nd_f_mes0bar,4)], [round(nd_prec1bar,4),round(nd_rap1bar,4),round(nd_f_mes1bar,4)]]
    
    return res
                                          
def get_scores_switch_time_alignment(time_alignment, segments, bars, ref_tab):
    """
    Get scores from segments, where the tolerance may be expressed in seconds (absolute time) or in bars (barwise aligned).
    """
    if type(ref_tab[0][0]) != np.ndarray: # ref_tab consist in the references, and should be nested in an array (for consistency).
        ref_tab = [ref_tab]

    # Tolerance in absolute time
    if time_alignment in ["s", "second", "seconds"]:
        return get_scores_in_time_from_barwise_segments(segments, bars, ref_tab)

    # Tolerance barwise aligned
    elif time_alignment in ["b", "bars", "bar", "barwise"]:
        return get_scores_in_bars_from_barwise_segments(segments, bars, ref_tab)
            
    else:
        raise NotImplementedError(f"Time alignment parameter {time_alignment} not understood")
    



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="as_seg",
    version="0.1.11",
    author="Marmoret Axel",
    author_email="axel.marmoret@imt-atlantique.fr",
    description="Package for the segmentation of autosimilarity matrices. This version is related to a stable vesion on PyPi, for installation in MSAF.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.imt-atlantique.fr/a23marmo/autosimilarity_segmentation",
    packages=setuptools.find_packages(),
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Multimedia :: Sound/Audio :: Analysis",
        "Programming Language :: Python :: 3.8"
    ],
    license='BSD',
    install_requires=[
        'base_audio',
        'librosa >= 0.10',
        'madmom',# @ git+https://github.com/CPJKU/madmom',
        'matplotlib >= 1.5',
        'mir_eval',
        'mirdata >= 0.3.3',
        'smart_open', # For mirdata, not installed by default, may be fixed in future release
        'numpy >= 1.8.0',
        'pandas',
        'scikit-learn >= 0.17.0',
        'scipy >= 0.13.0',
        'tensorly >= 0.5.1',
        'IPython'
    ]
)



"""
This module contains the dataloaders for the main datasets in MSA: SALAMI, RWCPOP and Beatles dataset.

It loads the data, but also computes the barwise TF matrix and the bars from the audio files.
When the barwise TF matrix and bars are computed, they are saved in a cache folder to avoid recomputing them.
"""

import mirdata
import librosa
import as_seg.model.signal_to_spectrogram as signal_to_spectrogram
import pathlib
import shutil
import numpy as np
import warnings

import as_seg

eps = 1e-10

class BaseDataloader():
    def __init__(self, feature, cache_path = None, sr=44100, hop_length = 32, subdivision = 96, verbose = False):
        """
        Constructor of the BaseDataloader class.

        Parameters
        ----------
        feature : string
            The feature to compute the spectrogram. Must be a valid feature name.
        cache_path : string
            The path where to save the computed barwise TF matrices and bars. If None, the cache is not used.
            The default is None.
        sr : int
            The sampling rate of the audio files.
            The default is 44100.
        hop_length : int
            The hop length of the spectrogram.
            The default is 32.
        subdivision : int
            The number of subdivisions of a bar.
            The default is 96.
        verbose : bool
            If True, print some information about the cache.
            The default is False
        """
        self.cache_path = cache_path
        self.verbose = verbose

        self.feature_object = signal_to_spectrogram.FeatureObject(sr, feature, hop_length)

        # For barwise or beatwise processing
        self.subdivision = subdivision

    def __getitem__(self, index):
        """
        Return the data of the index-th track.
        """
        raise NotImplementedError("This method should be implemented in the child class") from None

    def __len__(self):
        """
        Return the number of tracks in the dataset.
        """
        raise NotImplementedError("This method should be implemented in the child class") from None

    def get_spectrogram(self, signal): # The spectrogram is not saved in the cache because it is too large in general
        """
        Returns the spectrogram, from the signal of a song.
        """
        return self.feature_object.get_spectrogram(signal)
    
    def get_bars(self, audio_path, index = None):
        """
        Return the bars of the song.
        They are computed from the audio file.
        If the cache is used, the bars are saved in the cache.
        An identifier of the song should be provided to save the bars in the cache.
        """
        def _compute_bars(): # Define the function to compute the bars
            return as_seg.data_manipulation.get_bars_from_audio(audio_path)

        # If a cache is set
        if self.cache_path is not None:

            # No identifier is provided for this song, hence it cannot be saved in the cache
            if index is None:
                warnings.warn("No index provided for the cache, the cache will be ignored")
            
            # An identifier is provided
            else:
                dir_save_bars_path = f"{self.cache_path}/bars"
                
                # Tries to load the bars from the cache
                try:
                    bars = np.load(f"{dir_save_bars_path}/{index}.npy", allow_pickle=True)
                    if self.verbose:
                        print("Using cached bars.")
                
                # If the file is not found, the bars are computed and saved in the cache
                except FileNotFoundError:
                    bars = _compute_bars() # Compute the bars

                    # Save the bars in the cache
                    pathlib.Path(dir_save_bars_path).mkdir(parents=True, exist_ok=True)
                    np.save(f"{dir_save_bars_path}/{index}.npy", bars)
                
                # Return the bars
                return bars
        # No cache is set, the bars are computed and returned
        return _compute_bars()

    def get_barwise_tf_matrix(self, track_path, bars, index = None):
        """
        Return the barwise TF matrix of the song.
        It is computed from the signal of the song and the bars.
        If the cache is used, the barwise TF matrix is saved in the cache.
        An identifier of the song should be provided to save the barwise TF matrix in the cache.
        """
        def _compute_barwise_tf_matrix(): # Define the function to compute the barwise TF matrix
            # Load the signal of the song
            sig, _ = librosa.load(track_path, sr=self.sr, mono=True) #torchaudio.load(track.audio_path)
            # Compute the spectrogram
            spectrogram = self.get_spectrogram(sig)
            return as_seg.barwise_input.barwise_TF_matrix(spectrogram, bars, self.hop_length/self.sr, self.subdivision) + eps
        
        # If a cache is set 
        if self.cache_path is not None:
            # No identifier is provided for this song, hence it cannot be saved in the cache
            if index is None:
                warnings.warn("No index provided for the cache, the cache will be ignored")
            
            # An identifier is provided
            else:
                cache_file_name = f"{index}_{self.feature}_subdiv{self.subdivision}"
                dir_save_barwise_tf_path = f"{self.cache_path}/barwise_tf_matrix"
                
                # Tries to load the barwise TF matrix from the cache
                try:
                    barwise_tf_matrix = np.load(f"{dir_save_barwise_tf_path}/{cache_file_name}.npy", allow_pickle=True)
                    if self.verbose:
                        print("Using cached Barwise TF matrix.")
                
                # If the file is not found, the barwise TF matrix is computed and saved in the cache
                except FileNotFoundError:
                    barwise_tf_matrix = _compute_barwise_tf_matrix() # Compute the barwise TF matrix

                    # Save the barwise TF matrix in the cache
                    pathlib.Path(dir_save_barwise_tf_path).mkdir(parents=True, exist_ok=True)
                    np.save(f"{dir_save_barwise_tf_path}/{cache_file_name}.npy", barwise_tf_matrix)
                
                # Return the barwise TF matrix
                return barwise_tf_matrix
            
        # No cache is set, the barwise TF matrix is computed and returned
        return _compute_barwise_tf_matrix()

    def save_segments(self, segments, name):
        """
        Save the segments of a song in the original folder.
        Important for reproducibility.
        """
        # mirdata_segments = mirdata.annotations.SectionData(intervals=segments, interval_unit="s")
        # jams_segments = mirdata.jams_utils.sections_to_jams(mirdata_segments)
        dir_save_path = f"{self.data_path}/estimations/segments/{self.dataset_name.lower()}"
        pathlib.Path(dir_save_path).mkdir(parents=True, exist_ok=True)
        np.save(f"{dir_save_path}/{name}.npy", segments)

    def score_flat_segmentation(self, segments, annotations):
        """
        Compute the score of a flat segmentation.
        """
        close_tolerance = as_seg.data_manipulation.compute_score_of_segmentation(annotations, segments, window_length=0.5)
        large_tolerance = as_seg.data_manipulation.compute_score_of_segmentation(annotations, segments, window_length=3)
        return close_tolerance, large_tolerance
    
    def segments_from_bar_to_seconds(self, segments, bars):
        """
        Convert the segments from bars to seconds. Wrapper for the function in data_manipulation.
        """
        # May be useful, if ever.
        return as_seg.data_manipulation.segments_from_bar_to_time(segments, bars)

class RWCPopDataloader(BaseDataloader):
    """
    Dataloader for the RWC Pop dataset.
    """
    def __init__(self, path, feature, cache_path = None, download=False, sr=44100, hop_length = 32, subdivision = 96):
        """
        Constructor of the RWCPopDataloader class.

        Parameters
        ----------
        Same then for BaseDataloader, with the addition of:

        path : string
            The path to the dataset.
        download : bool
            If True, download the dataset using mirdata.
            The default is False.
        """
        super().__init__(feature=feature, cache_path=cache_path, sr=sr, hop_length=hop_length, subdivision=subdivision)
        self.data_path = path
        rwcpop = mirdata.initialize('rwc_popular', data_home = path)
        if download:
            rwcpop.download()            
        self.all_tracks = rwcpop.load_tracks()
        self.indexes = rwcpop.track_ids

        self.dataset_name = "RWCPop"

    def __getitem__(self, index):
        """
        Return the data of the index-th track.
        """
        track_id = self.indexes[index]
        track = self.all_tracks[track_id]

        # Compute the bars
        bars = self.get_bars(track.audio_path, index=track_id)

        # Compute the barwise TF matrix
        barwise_tf_matrix = self.get_barwise_tf_matrix(track.audio_path, bars, index=track_id)

        # Get the annotationks
        annotations_intervals = track.sections.intervals

        # Return the the bars, the barwise TF matrix and the annotations
        return track_id, bars, barwise_tf_matrix, annotations_intervals
    
    def __len__(self):
        """
        Return the number of tracks in the dataset.
        """
        return len(self.indexes)
    
    def get_track_of_id(self, track_id):
        """
        Return the data of the track with the given track_id.
        """
        index = self.indexes.index(track_id)
        return self.__getitem__[index]
    
    def format_dataset(self, path_audio_files):
        """
        Format the dataset from the mirdata way of downloading information to the way of the dataloader.
        """
        # Copy audio files to the right location.
        # Suppose that the audio files are all in the same folder
        for track_num in range(len(self.all_tracks)):
            track_idx = self.indexes[track_num]
            song_file_name = self.all_tracks[track_idx].audio_path.split('/')[-1]
            src = f"{path_audio_files}/{song_file_name}" # May change depending on your file structure
            dest = self.all_tracks[track_idx].audio_path
            pathlib.Path(dest).parent.absolute().mkdir(parents=True, exist_ok=True)
            shutil.copy(src, dest)

class SALAMIDataloader(BaseDataloader):
    """
    Dataloader for the SALAMI dataset.
    """
    def __init__(self, path, feature, cache_path = None, download=False, subset = None, sr=44100, hop_length = 32, subdivision = 96):
        """
        Constructor of the SALAMIDataloader class.

        Parameters
        ----------  
        Same then for BaseDataloader, with the addition of:

        path : string
            The path to the dataset.
        download : bool
            If True, download the dataset using mirdata.
            The default is False.
        subset : string
            The subset of the dataset to use. Can be "train", "test" or "debug".
        """
        super().__init__(feature=feature, cache_path=cache_path, sr=sr, hop_length=hop_length, subdivision=subdivision)
        
        self.dataset_name = "SALAMI"

        self.data_path = path
        salami = mirdata.initialize('salami', data_home = path)
        if download:
            salami.download()            
        self.all_tracks = salami.load_tracks()
        self.indexes = salami.track_ids

        self.subset = subset
        if subset is not None:
            train_indexes, test_indexes = self.split_training_test()
            if subset == "train":
                self.indexes = train_indexes
            elif subset == "test":
                self.indexes = test_indexes
            elif subset == "debug":
                self.indexes = test_indexes[:4]
            else:
                raise ValueError("Subset should be either 'train' or 'test'")


    def __getitem__(self, index):
        """
        Return the data of the index-th track.
        """
        # Parsing through files ordered with self.indexes
        track_id = self.indexes[index]
        track = self.all_tracks[track_id]

        try:           
            # Compute the bars
            bars = self.get_bars(track.audio_path, index=track_id)

            # Compute the barwise TF matrix
            barwise_tf_matrix = self.get_barwise_tf_matrix(track.audio_path, bars, index=track_id)

            # Get the annotations
            dict_annotations = self.get_annotations(track)

            # Return the the bars, the barwise TF matrix and the annotations
            return track_id, bars, barwise_tf_matrix, dict_annotations
    
        except FileNotFoundError:
            print(f'{track_id} not found.')
            return track_id, None, None, None
            # raise FileNotFoundError(f"Song {track_id} not found, normal ?") from None
    
    def __len__(self):
        """
        Return the number of tracks in the dataset, and in particular the number of tracks in the subset.
        """
        # To handle the fact that indexes are updated with the subset
        return len(self.indexes)

    def get_track_of_id(self, track_id):
        """
        Return the data of the track with the given track_id.
        """
        try:
            index = self.indexes.index(track_id)
        except ValueError:
            try:
                index = self.indexes.index(str(track_id))
            except ValueError:
                raise ValueError(f"Track {track_id} not found in the dataset") from None
        return self.__getitem__(index)

    def get_annotations(self, track):
        """
        Return the annotations of the track, in the form of a dict.
        It returns the annotations of the first annotator, and if available, the annotations of the second annotator.
        It returns both levels of annotations (upper and lower) for each annotator.
        """
        dict_annotations = {}
        try: 
            # Trying to get the first annotator
            dict_annotations["upper_level_annotations"] = track.sections_annotator_1_uppercase.intervals
            dict_annotations["lower_level_annotations"] = track.sections_annotator_1_lowercase.intervals
            try: # Trying to load the second annotator
                dict_annotations["upper_level_annotations_2"] = track.sections_annotator_2_uppercase.intervals
                dict_annotations["lower_level_annotations_2"] = track.sections_annotator_2_lowercase.intervals
                dict_annotations["annot_number"]  = 2
            except AttributeError: # Only the first annotator was loaded
                dict_annotations["annot_number"]  = 1
        except AttributeError:
            try:
                # Trying to get the second annotator (no first one)
                dict_annotations["upper_level_annotations"] = track.sections_annotator_2_uppercase.intervals
                dict_annotations["lower_level_annotations"] = track.sections_annotator_2_lowercase.intervals
                dict_annotations["annot_number"]  = 1
            except AttributeError:
                raise AttributeError(f"No annotations found for {track.track_id}")
        
        return dict_annotations
    
    def get_this_set_annotations(self, dict_annotations, annotation_level = "upper", annotator = 1):
        """
        Return a particular set of annotations from all the annotations.
        """
        if annotator == 1:
            if annotation_level == "upper":
                annotations = dict_annotations["upper_level_annotations"]
            elif annotation_level == "lower":
                annotations = dict_annotations["lower_level_annotations"]
            else:
                raise ValueError("Invalid annotation level")
        elif annotator == 2:
            assert dict_annotations["annot_number"] == 2, "No second annotator found."
            if annotation_level == "upper":
                annotations = dict_annotations["upper_level_annotations"]
            elif annotation_level == "lower":
                annotations = dict_annotations["lower_level_annotations"]
            else:
                raise ValueError("Invalid annotation level")
        # elif annotator == "both":
        #     assert dict_annotations["annot_number"] == 2, "No second annotator found."
        #     annotations = dict_annotations["upper_level_annotations"] + dict_annotations["upper_level_annotations_2"]
        else:
            raise ValueError("Invalid annotator number")
        return annotations

    def split_training_test(self):
        """
        Split the dataset in training and test set.
        """
        indexes_train = []
        indexes_test = []
        for track_id in self.indexes:
            track = self.all_tracks[track_id]
            try:
                track.sections_annotator_1_uppercase.intervals
                track.sections_annotator_2_uppercase.intervals
                indexes_test.append(track_id)
            except AttributeError:
                indexes_train.append(track_id)
        return indexes_train, indexes_test
    
    def score_flat_segmentation(self, segments, dict_annotations, annotation_level = "upper", annotator = 1):
        """
        Score a flat segmentation.
        """
        if annotator == "both":
            assert dict_annotations["annot_number"] == 2, "No second annotator found."
            score_annot_1 = self.score_flat_segmentation(segments, dict_annotations, annotation_level = annotation_level, annotator = 1)
            score_annot_2 = self.score_flat_segmentation(segments, dict_annotations, annotation_level = annotation_level, annotator = 2)
            return score_annot_1, score_annot_2
        
        annotations = self.get_this_set_annotations(dict_annotations, annotation_level = annotation_level, annotator = annotator)
        return super().score_flat_segmentation(segments, annotations)
        
    def score_flat_segmentation_twolevel(self, segments_upper_level, segments_lower_level, dict_annotations, annotator = 1):
        """
        Score a flat segmentation at both levels of annotations.
        """
        score_upper_level = self.score_flat_segmentation(segments_upper_level, dict_annotations, annotation_level = "upper", annotator = annotator)
        score_lower_level = self.score_flat_segmentation(segments_lower_level, dict_annotations, annotation_level = "lower", annotator = annotator)
        return score_upper_level, score_lower_level
    
    def score_flat_segmentation_twolevel_best_of_several(self, list_segments_upper_level, list_segments_lower_level, dict_annotations, annotator = 1):
        """
        Score a flat segmentation at both levels of annotations, and return the best score from the different annotators.
        """
        assert annotator != "both", "Not implemented yet"
        stack_upper_scores = -np.inf * np.ones((len(list_segments_upper_level),2,3))
        for idx, segments in enumerate(list_segments_upper_level):
            stack_upper_scores[idx] = self.score_flat_segmentation(segments, dict_annotations, annotation_level = "upper", annotator = annotator)
        idx_close = np.argmax(stack_upper_scores[:,0,2]) # Selecting best f measure at 0.5s
        idx_large = np.argmax(stack_upper_scores[:,1,2]) # Selecting best f measure at 3s
        score_upper_level = (stack_upper_scores[idx_close,0,:], stack_upper_scores[idx_large,1,:])

        stack_lower_scores = -np.inf * np.ones((len(list_segments_lower_level),2,3))
        for idx, segments in enumerate(list_segments_lower_level):
            stack_lower_scores[idx] = self.score_flat_segmentation(segments, dict_annotations, annotation_level = "lower", annotator = annotator)
        idx_close = np.argmax(stack_lower_scores[:,0,2]) # Selecting best f measure at 0.5s
        idx_large = np.argmax(stack_lower_scores[:,1,2]) # Selecting best f measure at 3s
        score_lower_level = (stack_lower_scores[idx_close,0,:], stack_lower_scores[idx_large,1,:])

        return score_upper_level, score_lower_level

    def get_sizes_of_annotated_segments(self, annotation_level = "upper", annotator = 1, plot = False):
        """
        Return the lengths of the annotated segments.
        """
        lengths = []
        for track_id in self.indexes:
            track = self.all_tracks[track_id]

            try:           
                # Compute the bars
                bars = self.get_bars(track.audio_path, index=track_id)

                # Get the annotations
                dict_annotations = self.get_annotations(track)

                annotations = self.get_this_set_annotations(dict_annotations, annotation_level = annotation_level, annotator = annotator)

                barwise_annot = as_seg.data_manipulation.frontiers_from_time_to_bar(np.array(annotations)[:,1], bars) # Convert the annotations from time to bar
                for i in range(len(barwise_annot) - 1):
                    lengths.append(barwise_annot[i+1] - barwise_annot[i]) # Store the length of the annotated segment
        
            except FileNotFoundError:
                print(f'{track_id} not found.')
                # raise FileNotFoundError(f"Song {track_id} not found, normal ?") from None

        if plot:
            as_seg.model.common_plot.plot_lenghts_hist(lengths)
        return lengths
        
    # def format_dataset(self, path_audio_files): # TODO
        # # Copy audio files to the right location.
        # # Suppose that the audio files are all in the same folder
        # for track_num in range(len(self.all_tracks)):
        #     track_idx = self.indexes[track_num]
        #     song_file_name = self.all_tracks[track_idx].audio_path.split('/')[-1]
        #     src = f"{path_audio_files}/{song_file_name}" # May change depending on your file structure
        #     dest = self.all_tracks[track_idx].audio_path
        #     pathlib.Path(dest).parent.absolute().mkdir(parents=True, exist_ok=True)
        #     shutil.copy(src, dest)
    

class BeatlesDataloader(BaseDataloader):
    """
    Dataloader for the Beatles dataset.
    """
    def __init__(self, path, feature, cache_path = None, download=False, sr=44100, hop_length = 32, subdivision = 96):
        """
        Constructor of the BeatlesDataloader class.

        Parameters
        ----------
        Same then for BaseDataloader, with the addition of:

        path : string
            The path to the dataset.
        download : bool
            If True, download the dataset using mirdata.
            The default is False
        """
        super().__init__(feature, cache_path, sr, hop_length, subdivision)
        self.data_path = path
        beatles = mirdata.initialize('beatles', data_home = path)
        if download:
            beatles.download()            
        self.all_tracks = beatles.load_tracks()
        self.indexes = beatles.track_ids

        self.dataset_name = "Beatles"

    def __getitem__(self, index):
        """
        Return the data of the index-th track.
        """
        track_id = self.indexes[index]
        track = self.all_tracks[track_id]

        # Compute the bars
        bars = self.get_bars(track.audio_path, index=track_id)

        # Compute the barwise TF matrix
        barwise_tf_matrix = self.get_barwise_tf_matrix(track.audio_path, bars, index=track_id)

        # Get the annotationks
        annotations_intervals = track.sections.intervals

        # Return the the bars, the barwise TF matrix and the annotations
        return track_id, bars, barwise_tf_matrix, annotations_intervals
    
    def __len__(self):
        """
        Return the number of tracks in the dataset.
        """
        return len(self.all_tracks)
    
    def get_track_of_id(self, track_id):
        """
        Return the data of the track with the given track_id.
        """
        try:
            index = self.indexes.index(track_id)
        except ValueError:
            try:
                index = self.indexes.index(str(track_id))
            except ValueError:
                raise ValueError(f"Track {track_id} not found in the dataset") from None
        return self.__getitem__(index)

if __name__ == "__main__":
    # rwcpop = RWCPopDataloader('/home/a23marmo/datasets/rwcpop', feature = "mel", cache_path = "/home/a23marmo/Bureau/data_persisted/rwcpop")
    # # rwcpop.format_dataset('/home/a23marmo/Bureau/Audio samples/rwcpop/Entire RWC')
    # for spectrogram, bars, barwise_tf_matrix, track_id, annotations in rwcpop:
    #     print(spectrogram.shape, track_id)

    salami = SALAMIDataloader('/home/a23marmo/datasets/salami', feature = "mel", cache_path = "/home/a23marmo/Bureau/data_persisted/salami", subset = "train")

    for spectrogram, bars, barwise_tf_matrix, track_id, annotations in salami:
        try:
            print(track_id)
        except FileNotFoundError:
            print(f'{track_id} not found.')
